# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Base definitions for resource manager classes.

Resource manager classes exist for each resource type and are helper classes
that provide functionality common for the resource type.

Resource manager objects are not necessarily singleton objects, because they
have a scope of a certain set of resource objects. For example, the resource
manager object for LPARs exists once for each CPC managed by the HMC, and the
resource object scope of each LPAR manager object is the set of LPARs in that
CPC.
"""


import re
from datetime import datetime, timedelta
import time
import warnings
import threading
from nocasedict import NocaseDict

from ._logging import logged_api_call
from ._exceptions import NotFound, NoUniqueMatch, HTTPError
from ._utils import repr_list, matches_filters, divide_filter_args, \
    make_query_str, RC_LOGICAL_PARTITION, repr_obj_id

__all__ = ['BaseManager']


REGEXP_SPECIAL_CHAR = re.compile(r'[\^\$\.\+\*\?\(\)\[\]\{\}\|\\]')


class _NameUriCache:
    """
    A Name-URI cache, that caches the mapping between resource names and
    resource URIs. It supports looking up resource URIs by resource names.

    This class is used by the implementation of manager classes, and is not
    part of the external API.
    """

    def __init__(self, manager, timetolive, case_insensitive_names):
        """
        Parameters:

          manager (BaseManager): Manager that holds this Name-URI cache. The
            manager object is expected to have a ``list()`` method, which
            is used to list the resources of that manager, in order to
            fill this cache.

          timetolive (number): Time in seconds until the cache will invalidate
            itself automatically, since it was last invalidated.

          case_insensitive_names (bool): Controls whether the name of the
            resource is treated case insensitively.
        """
        self._manager = manager
        self._timetolive = timetolive
        self._dict_type = NocaseDict if case_insensitive_names else dict

        # The cached data, as a dictionary with:
        # Key (string): Name of a resource (unique within its parent resource).
        # Value (string): tuple(name, uri) where name is the original name
        # of the resource (important for resources with case-insensitive names)
        # and uri is the URI of the resource.
        self._uris = self._dict_type()

        # Point in time when the cache was last invalidated
        self._invalidated = datetime.now()

    def get(self, name):
        """
        Get the resource name and URI for a specified resource name as a
        tuple(name, uri).

        Note that for case-inensitive caches, it may be important to get back
        the original name, so both name and URI are returned.

        If an entry for the specified resource name does not exist in the
        Name-URI cache, the cache is refreshed from the HMC with all resources
        of the manager holding this cache.

        If an entry for the specified resource name still does not exist after
        that, ``NotFound`` is raised.
        """
        self.auto_invalidate()
        try:
            return self._uris[name]
        except KeyError:
            self.refresh()
            try:
                return self._uris[name]
            except KeyError:
                # pylint: disable=protected-access
                new_exc = NotFound(
                    {self._manager._name_prop: name}, self._manager)
                new_exc.__cause__ = None
                raise new_exc  # zhmcclient.NotFound

    def auto_invalidate(self):
        """
        Invalidate the cache if the current time is past the time to live.
        """
        current = datetime.now()
        if current > self._invalidated + timedelta(seconds=self._timetolive):
            self.invalidate()

    def invalidate(self):
        """
        Invalidate the cache.

        This empties the cache and sets the time of last invalidation to the
        current time.
        """
        self._uris = self._dict_type()
        self._invalidated = datetime.now()

    def refresh(self):
        """
        Refresh the Name-URI cache from the HMC.

        This is done by invalidating the cache, listing the resources of this
        manager from the HMC, and populating the cache with that information.
        """
        # pylint: disable=protected-access
        self.invalidate()
        full = not self._manager._list_has_name
        res_list = self._manager.list(full_properties=full)
        self.update_from(res_list)

    def update_from(self, res_list):
        """
        Update the Name-URI cache from the provided resource list.

        This is done by going through the resource list and updating any cache
        entries for non-empty resource names in that list. Other cache entries
        remain unchanged.
        """
        # pylint: disable=protected-access
        for res in res_list:
            # We access the properties dictionary, in order to make sure
            # we don't drive additional HMC interactions.
            name = res.properties.get(self._manager._name_prop, None)
            uri = res.properties.get(self._manager._uri_prop, None)
            self.update(name, uri)

    def update(self, name, uri):
        """
        Update or create the entry for the specified resource name in the
        Name-URI cache, and set it to the specified name and URI as a
        tuple(name, uri).

        If the specified name is `None` or the empty string, do nothing.
        """
        if name:
            self._uris[name] = (name, uri)

    def delete(self, name):
        """
        Delete the entry for the specified resource name from the Name-URI
        cache.

        If the specified name is `None` or the empty string, or if an entry for
        the specified name does not exist, do nothing.
        """
        if name:
            try:
                del self._uris[name]
            except KeyError:
                pass


class _ResourceList:
    """
    A list of resources, for use by resource manager objects for auto-updating.

    The resources in the list are the zhmcclient resource objects, organized by
    resource URI.

    This class is used by the implementation of manager classes, and is not
    part of the external API.
    """

    def __init__(self, manager):
        """
        Parameters:

          manager (BaseManager): Manager that holds this list of resources. The
            manager object is expected to have a ``list()`` method, which is
            used to list the resources of that manager.
        """
        self._manager = manager

        # Attributes that are updated under the lock
        self._lock = threading.RLock()
        self._resources = {}  # key: resource URI, value: resource obj
        self._needs_pull = True  # list() method needs to pull from HMC
        self._enabled = False  # Auto-updating of manager is enabled

    def __repr__(self):
        """
        Return a string with the state of this object, for debug purposes.
        """
        ret = (
            f"{repr_obj_id(self)} (\n"
            f"  _enabled={self._enabled!r},\n"
            f"  _resources(keys)={list(self._resources.keys())!r}\n"
            ")")
        return ret

    def enabled(self):
        """
        Return whether this list of resources is enabled.

        Return:
          bool: Indicates whether this list of resources is enabled.
        """
        return self._enabled

    def needs_pull(self):
        """
        Return whether this list of resources needs to be pulled from the HMC.
        """
        return self._needs_pull

    def enable(self):
        """
        Enable this list of resources, if currently disabled.

        When enabling, the session to which this manager belongs is subscribed
        for auto-updating if needed (see
        :meth:`~zhmcclient.Session.subscribe_auto_update`), the manager
        object is registered with the session's auto updater via
        :meth:`~zhmcclient.AutoUpdater.register_object`, and all resources
        of this manager object are retrieved using :meth:`list` in order to
        have the most current list of resources as a basis for the future
        auto-updating.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        if not self._enabled:
            session = self._manager.session
            session.subscribe_auto_update()
            session.auto_updater.register_object(self._manager)

            # The following list() call needs to pull from HMC. That is
            # happening because the resource list is still disabled at this
            # point.
            resource_list = self._manager.list()

            with self._lock:
                self._resources = {}
                for res_obj in resource_list:
                    self._resources[res_obj.uri] = res_obj
                self._needs_pull = False
                self._enabled = True

    def disable(self):
        """
        Disable this list of resources, if currently enabled.

        When disabling, the manager object is unregistered from the session's
        auto updater via
        :meth:`~zhmcclient.AutoUpdater.unregister_object`, and the session
        is unsubscribed from auto-updating if the auto updater has no more
        objects registered. Also, the list of resources is cleared.
        """
        if self._enabled:
            session = self._manager.session
            session.auto_updater.unregister_object(self._manager)
            if not session.auto_updater.has_objects():
                session.unsubscribe_auto_update()
            with self._lock:
                self._resources = {}
                self._needs_pull = True
                self._enabled = False

    def list(self):
        """
        Return a new list with the resource objects from this list of resources.
        """
        res_list = []
        with self._lock:
            for res_obj in self._resources.values():
                res_list.append(res_obj)
        return res_list

    def list_uris(self):
        """
        Return a new list with the resource URIs from this list of resources.
        """
        uri_list = []
        with self._lock:
            for res_uri in self._resources:
                uri_list.append(res_uri)
        return uri_list

    def add_list(self, resource_obj_list):
        """
        Add a new resource object list to this list of resources and mark
        it as no longer needing pull from the HMC.

        This method is called in list() to put resources pulled from the HMC
        into the list of resources. It should not be called by the user.
        """
        with self._lock:
            for res_obj in resource_obj_list:
                self._resources[res_obj.uri] = res_obj
            self._needs_pull = False

    def trigger_pull(self):
        """
        Trigger that resources need to be pulled from the HMC upon the next
        list() call of the manager object.

        This method is called when an inventory change notification indicates
        that a new object on the HMC has been created. It should not be called
        by the user.
        """
        with self._lock:
            self._needs_pull = True

    def remove(self, resource_uri):
        """
        Remove the item for a resource URI from this list of resources.

        If the resource URI is not in that list, do nothing.

        This method is called when an inventory change notification indicates
        that an object on the HMC has been deleted. It should not be called
        by the user.
        """
        with self._lock:
            try:
                del self._resources[resource_uri]
            except KeyError:
                pass


class BaseManager:
    """
    Abstract base class for manager classes (e.g.
    :class:`~zhmcclient.CpcManager`).

    It defines the interface for the derived manager classes, and implements
    methods that have a common implementation for the derived manager classes.

    Objects of derived manager classes should not be created by users of this
    package by simply instantiating them. Instead, such objects are created by
    this package as instance variables of :class:`~zhmcclient.Client` and
    other resource objects, e.g. :attr:`~zhmcclient.Client.cpcs`. For this
    reason, the `__init__()`  method of this class and of its derived manager
    classes are considered internal interfaces and their parameters are not
    documented and may change incompatibly.
    """

    def __init__(self, resource_class, class_name, session, parent, base_uri,
                 oid_prop, uri_prop, name_prop, query_props,
                 list_has_name=True, case_insensitive_names=False,
                 supports_properties=False):
        # This method intentionally has no docstring, because it is internal.
        #
        # Parameters:
        #   resource_class (class):
        #     Python class for the resources of this manager.
        #     Must not be `None`.
        #   class_name (string):
        #     Resource class name (e.g. 'cpc' for a CPC resource). Must
        #     be the value of the 'class' property of the resource.
        #     Must not be `None`.
        #   session (:class:`~zhmcclient.Session`):
        #     Session for this manager.
        #     Must not be `None`.
        #   parent (subclass of :class:`~zhmcclient.BaseResource`):
        #     Parent resource defining the scope for this manager.
        #     `None`, if the manager has no parent, i.e. when it manages
        #     top-level resources (e.g. CPC).
        #   base_uri (string):
        #     Base URI of the resources of this manager. The base URI has no
        #     trailing slash and becomes the resource URI by appending '/' and
        #     the value of the property specified in 'oid_prop'.
        #     Must not be `None`.
        #   oid_prop (string):
        #     Name of the resource property whose value is appended to the
        #     base URI to form the resource URI (e.g. 'object-id' or
        #     'element-id').
        #     Must not be `None`.
        #   uri_prop (string):
        #     Name of the resource property that is the canonical URI path of
        #     the resource (e.g. 'object-uri' or 'element-uri').
        #     Must not be `None`.
        #   name_prop (string):
        #     Name of the resource property that is the name of the resource
        #     (e.g. 'name').
        #     Must not be `None`.
        #   query_props (iterable of strings):
        #     List of names of resource properties that are supported as filter
        #     query parameters in HMC list operations for this type of resource
        #     (i.e. for server-side filtering).
        #     May be `None`.
        #     If the support for a resource property changes within the set of
        #     HMC versions that support this type of resource, this list must
        #     represent the version of the HMC this session is connected to.
        #   list_has_name (bool):
        #     Indicates whether the list() method for the resource populates
        #     the name property (i.e. name_prop). For example, for NICs the
        #     list() method returns minimalistic Nic objects without name.
        #   case_insensitive_names (bool):
        #     Indicates whether the name of the resource is treated case
        #     insensitively.
        #   supports_properties (bool):
        #     Indicates whether the Get Properties operation for this type of
        #     resource supports the 'properties' query parameter in the latest
        #     released version of the HMC.

        # We want to surface precondition violations as early as possible,
        # so we test those that are not surfaced through the init code:
        assert resource_class is not None
        assert class_name is not None
        assert session is not None
        assert base_uri is not None
        assert oid_prop is not None
        assert uri_prop is not None
        assert name_prop is not None

        self._resource_class = resource_class
        self._class_name = class_name
        self._uri = None
        self._session = session
        self._parent = parent
        self._base_uri = base_uri
        self._oid_prop = oid_prop
        self._uri_prop = uri_prop
        self._name_prop = name_prop
        self._query_props = query_props
        self._list_has_name = list_has_name
        self._case_insensitive_names = case_insensitive_names
        self._supports_properties = supports_properties

        self._resource_list = _ResourceList(self)
        self._name_uri_cache = _NameUriCache(
            self, session.retry_timeout_config.name_uri_cache_timetolive,
            case_insensitive_names)

    def __repr__(self):
        """
        Return a string with the state of this manager object, for debug
        purposes.
        """
        ret = (
            f"{repr_obj_id(self)} (\n"
            f"  _resource_class={self._resource_class!r},\n"
            f"  _class_name={self._class_name!r},\n"
            f"  _uri={self._uri!r},\n"
            f"  _session={repr_obj_id(self._session)},\n"
            f"  _parent={repr_obj_id(self._parent)},\n"
            f"  _base_uri={self._base_uri!r},\n"
            f"  _oid_prop={self._oid_prop!r},\n"
            f"  _uri_prop={self._uri_prop!r},\n"
            f"  _name_prop={self._name_prop!r},\n"
            f"  _query_props={repr_list(self._query_props, indent=2)},\n"
            f"  _list_has_name={self._list_has_name!r},\n"
            f"  _case_insensitive_names={self._case_insensitive_names!r},\n"
            f"  _supports_properties={self._supports_properties!r},\n"
            f"  _resource_list={self._resource_list!r},\n"
            f"  _name_uri_cache={self._name_uri_cache!r}\n"
            ")")
        return ret

    def invalidate_cache(self):
        """
        Invalidate the Name-URI cache of this manager.

        The zhmcclient maintains a Name-URI cache in each manager object, which
        caches the mappings between resource URIs and resource names, to speed
        up certain zhmcclient methods.

        The Name-URI cache is properly updated during changes on the resource
        name (e.g. via :meth:`~zhmcclient.Partition.update_properties`) or
        changes on the resource URI (e.g. via resource creation or deletion),
        if these changes are performed through the same Python manager object.

        However, changes performed through a different manager object (e.g.
        because a different session, client or parent resource object was
        used), or changes performed in a different Python process, or changes
        performed via other means than the zhmcclient library (e.g. directly on
        the HMC) will not automatically update the Name-URI cache of this
        manager.

        In cases where the resource name or resource URI are effected by such
        changes, the Name-URI cache can be manually invalidated by the user,
        using this method.

        Note that the Name-URI cache automatically invalidates itself after a
        certain time since the last invalidation. That auto invalidation time
        can be configured using the
        :attr:`~zhmcclient.RetryTimeoutConfig.name_uri_cache_timetolive`
        attribute of the :class:`~zhmcclient.RetryTimeoutConfig` class.
        """
        self._name_uri_cache.invalidate()

    def _try_optimized_lookup(self, filter_args):
        """
        Try to find a resource in an optimized way when the filter arguments
        allow for that.

        The following cases of optimized filtering are supported:

        - A single filter on the resource name, that does not use regular
          expression matching.
        - A single filter on the resource object/element ID, that does not use
          regular expression matching.

        Returns `None` if the filter arguments do not meet these optimization
        criteria, or if they do but no resource was found.

        Parameters:

          filter_args (dict):
            Filter arguments. For details, see :ref:`Filtering`.

        Returns:

          resource object, or `None` if the the filter arguments did not meet
            the optimization criteria, or if they did but no resource was found.
        """
        if filter_args is None or len(filter_args) != 1:
            return None

        if self._name_prop in filter_args:

            name_match = filter_args[self._name_prop]
            if not isinstance(name_match, str) or \
                    REGEXP_SPECIAL_CHAR.search(name_match):
                return None

            try:
                resource_obj = self.find_by_name(name_match)
            except NotFound:
                return None

            return resource_obj

        if self._oid_prop in filter_args:

            oid_match = filter_args[self._oid_prop]
            if not isinstance(oid_match, str) or \
                    REGEXP_SPECIAL_CHAR.search(oid_match):
                return None

            # Construct the resource URI from the filter property
            # and issue a Get <Resource> Properties on that URI
            uri = self._base_uri + '/' + oid_match
            try:
                props = self.session.get(uri)
            except HTTPError as exc:
                if exc.http_status == 404 and exc.reason == 1:
                    # No such resource
                    return None
                raise

            resource_obj = self.resource_class(
                manager=self,
                uri=props[self._uri_prop],
                name=props.get(self._name_prop, None),
                properties=props)

            # pylint: disable=protected-access
            resource_obj._full_properties = True

            return resource_obj

        return None

    @property
    def resource_class(self):
        """
        The Python class of the parent resource of this manager.
        """
        return self._resource_class

    @property
    def class_name(self):
        """
        The resource class name
        """
        return self._class_name

    @property
    def name_prop(self):
        """
        The name of the resource property indicating the resource name
        """
        return self._name_prop

    @property
    def session(self):
        """
        :class:`~zhmcclient.Session`:
          Session with the HMC.
        """
        assert self._session is not None, (
            f"{self.__class__.__name__}.session: No session set (in top-level "
            "resource manager class?)")
        return self._session

    @property
    def parent(self):
        """
        Subclass of :class:`~zhmcclient.BaseResource`:
          Parent resource defining the scope for this manager.

          `None`, if the manager has no parent, i.e. when it manages top-level
          resources.
        """
        return self._parent

    @property
    def uri(self):
        """
        string: The canonical URI path of the manager. Will not be `None`.

        This URI uniquely identifies the list of HMC resources in scope of the
        manager, consistent with how the canonical URI path of a resource
        identifies the HMC resource.

        The format of this URI is undocumented.

        This URI is used in the implementation of auto-updated manager objects,
        it does not have any meaning on the HMC, and there should be no need
        for users to use it.
        """
        if self._uri is None:
            if self._parent:
                parent_uri = self._parent.uri
            else:
                parent_uri = '/'  # top-level resource
            self._uri = f'{parent_uri}#{self._class_name}'
        return self._uri

    @property
    def case_insensitive_names(self):
        """
        :class:`py:bool`:
          Indicates whether the names of the resources are treated case
          insensitively.
        """
        return self._case_insensitive_names

    @property
    def supports_properties(self):
        """
        :class:`py:bool`:
          Indicates whether the "Get Properties" operation for this type of
          resource supports the 'properties' query parameter in the latest
          released version of the HMC.
        """
        return self._supports_properties

    def _list_with_operation(
            self, list_uri, result_prop, full_properties, filter_args,
            additional_properties):
        """
        List resource objects by using a List operation.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        Parameters:

          list_uri (string):
            Canonical URI for the list operation.

          result_prop (string):
            Name of property in result of list operation that contains
            the resource list.

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

            Must be `None` for resource types whose List operation does not
            support the 'additional-properties' query parameter.

        Returns:

          : A list of zhmcclient resource objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        resource_obj_list = []
        if self.auto_update_enabled() and not self.auto_update_needs_pull():
            for resource_obj in self.list_resources_local():
                if matches_filters(resource_obj, filter_args):
                    resource_obj_list.append(resource_obj)
        else:
            query_parms, client_filters = divide_filter_args(
                self._query_props, filter_args)
            if additional_properties:
                ap_parm = \
                    f"additional-properties={','.join(additional_properties)}"
                query_parms.append(ap_parm)
            query_parms_str = make_query_str(query_parms)
            uri = f'{list_uri}{query_parms_str}'

            try:
                result = self.session.get(uri)
            except HTTPError as exc:
                if self.class_name == RC_LOGICAL_PARTITION and \
                        exc.http_status == 404 and exc.reason == 1:
                    # Unlike other list operations, "List Logical Partitions
                    # of CPC" fails with 404.1 "ERROR: found no Images" if
                    # no LPAR matches the filters in the query parms.
                    result = []
                else:
                    raise

            if result:
                props_list = result[result_prop]
                if full_properties:
                    resource_obj_list.extend(
                        self._get_properties_bulk(
                            props_list, client_filters))
                else:
                    for props in props_list:
                        resource_obj = self.resource_class(
                            manager=self,
                            uri=props[self._uri_prop],
                            name=props.get(self._name_prop, None),
                            properties=props)
                        if matches_filters(resource_obj, client_filters):
                            resource_obj_list.append(resource_obj)

            self.add_resources_local(resource_obj_list)

        self._name_uri_cache.update_from(resource_obj_list)
        return resource_obj_list

    def _get_properties_bulk(self, props_list, client_filters):
        """
        Get resource properties using a bulk operation.

        Parameters:

          props_list (list of dict):
            List of resource properties from List operation.
            Must contain the resource URIs.

            These properties will appear in the returned zhmcclient resource
            objects, updated by the actual property values from the HMC.
            This allows adding properties to those returned by the HMC.

          client_filters (dict):
            Filter arguments to be applied on the client side after the
            resource properties have been retrieved.
            `None` causes no client filtering to happen.

        Returns:

          : A list of zhmcclient resource objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        resource_obj_list = []

        bulk_reqs = []
        req_by_id = {}
        req_id = 0
        for props in props_list:
            req_id += 1
            req_id_str = str(req_id)
            req = {
                'method': 'GET',
                'uri': props[self._uri_prop],
                'id': req_id_str,
            }
            bulk_reqs.append(req)
            req_by_id[req_id_str] = (req, props)

        if bulk_reqs:
            uri = '/api/services/aggregation/submit'
            # 10 threads is the supported maximum
            threads = min(10, round(len(props_list) / 2 + 0.51))
            body = {
                'requests': bulk_reqs,
                'threads': threads,
            }
            result = self.session.post(uri, body=body)
            for res in result:
                req_id = res['id']
                req, props = req_by_id[req_id]

                # We first use the properties from the props_list parameter,
                # and then update that with the properties returned from the
                # HMC:
                resource_props = dict(props)
                resource_props.update(res['body'])

                if res['status'] != 200:
                    # Similar to the non-full case: The first
                    # error raises an exception.
                    raise HTTPError(resource_props)

                resource_obj = self.resource_class(
                    manager=self,
                    uri=req['uri'],
                    name=resource_props.get(self._name_prop, None),
                    properties=resource_props)

                # pylint: disable=protected-access
                with resource_obj._property_lock:
                    resource_obj._properties = dict(resource_props)
                    resource_obj._properties_timestamp = int(time.time())
                    resource_obj._full_properties = True
                # pylint: enable=protected-access

                if matches_filters(resource_obj, client_filters):
                    resource_obj_list.append(resource_obj)

        return resource_obj_list

    def _list_with_parent_array(
            self, parent_obj, uris_prop, full_properties, filter_args):
        """
        List resource objects by using an array of URIs in the parent object.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, the corresponding array property for this resource in the
          parent object is used to list the resources, and the provided filter
          arguments are applied.

        Parameters:

          parent_obj (zhmcclient.BaseResource):
            The parent object.

          uris_prop (string):
            Name of the array property with URIs in the parent object.

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of zhmcclient resource objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        resource_obj_list = []
        if self.auto_update_enabled() and not self.auto_update_needs_pull():
            for resource_obj in self.list_resources_local():
                if matches_filters(resource_obj, filter_args):
                    resource_obj_list.append(resource_obj)
        else:
            uris = parent_obj.get_property(uris_prop)
            if uris:
                for uri in uris:

                    resource_obj = self.resource_class(
                        manager=self,
                        uri=uri,
                        name=None,
                        properties=None)

                    if matches_filters(resource_obj, filter_args):
                        resource_obj_list.append(resource_obj)
                        if full_properties:
                            resource_obj.pull_full_properties()

            self.add_resources_local(resource_obj_list)

        self._name_uri_cache.update_from(resource_obj_list)
        return resource_obj_list

    def auto_update_enabled(self):
        """
        Return whether :ref:`auto-updating` is
        currently enabled for the manager object.

        Return:
          bool: Indicates whether auto-update is enabled.
        """
        return self._resource_list.enabled()

    def auto_update_needs_pull(self):
        """
        Return whether there is a need to pull the resources from the HMC, in
        the list() method.

        This method is called in the list() method. It should not be called
        by the user.
        """
        return self._resource_list.needs_pull()

    def enable_auto_update(self):
        """
        Enable :ref:`auto-updating` for this manager object, if currently
        disabled.

        When enabling auto-update, the session to which this manager belongs is
        subscribed for auto-updating if needed (see
        :meth:`~zhmcclient.Session.subscribe_auto_update`), the manager
        object is registered with the session's auto updater via
        :meth:`~zhmcclient.AutoUpdater.register_object`, and all resources
        of this manager object are retrieved using :meth:`list` in order to
        have the most current list of resources as a basis for the future
        auto-updating.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        self._resource_list.enable()

    def disable_auto_update(self):
        """
        Disable :ref:`auto-updating` for this manager object, if currently
        enabled.

        When disabling auto-updating, the manager object is unregistered from
        the session's auto updater via
        :meth:`~zhmcclient.AutoUpdater.unregister_object`, and the session
        is unsubscribed from auto-updating if the auto updater has no more
        objects registered.
        """
        self._resource_list.disable()

    def auto_update_trigger_pull(self):
        """
        Trigger the need to pull the resources from the HMC, in the list()
        method.

        This method is called when an inventory change notification indicates
        that a new object on the HMC has been created. It should not be called
        by the user.
        """
        self._resource_list.trigger_pull()

    def add_resources_local(self, resource_obj_list):
        """
        Add a resource object to the local auto-updated list of resources.

        This method is called in list() to put resources pulled from the HMC
        into the list of resources. It should not be called by the user.
        """
        self._resource_list.add_list(resource_obj_list)

    def remove_resource_local(self, resource_uri):
        """
        Remove the resource object for a resource URI from the local
        auto-updated list of resources.

        If the resource URI is not in that list, do nothing.

        This method is called when an inventory change notification indicates
        that an object on the HMC has been deleted. It should not be called
        by the user.
        """
        self._resource_list.remove(resource_uri)

    def list_resources_local(self):
        """
        List the resource objects from the local auto-updated list of resources.

        This method is called by the list() methods of resource manager classes.
        It should not be called by the user.
        """
        return self._resource_list.list()

    def resource_object(self, uri_or_oid, props=None):
        """
        Return a minimalistic Python resource object for this resource class,
        that is scoped to this manager.

        This method is an internal helper function and is not normally called
        by users.

        The returned resource object will have the following minimal set of
        properties set automatically:

          * `object-uri` or `element-uri`
          * `object-id` or `element-id`
          * `parent`
          * `class`

        Additional properties for the Python resource object can be specified
        by the caller.

        Parameters:

            uri_or_oid (string): `object-uri` or `object-id` of the resource.

            props (dict): Property values in addition to the minimal list of
              properties that are set automatically (see above).

        Returns:

            Subclass of :class:`~zhmcclient.BaseResource`: A Python resource
            object for this resource class.
        """
        if uri_or_oid.startswith('/api/'):
            assert uri_or_oid[-1] != '/'
            uri = uri_or_oid
            oid = uri.split('/')[-1]
            # For the Console, we can predict the URI but not the OID.
            if oid == 'console':
                oid = None
        else:
            assert '/' not in uri_or_oid
            oid = uri_or_oid
            uri = f'{self._base_uri}/{oid}'
        res_props = {
            'parent': self.parent.uri if self.parent is not None else None,
            'class': self.class_name,
        }
        if oid:
            res_props[self._oid_prop] = oid
        name = None
        if props:
            res_props.update(props)
            try:
                name = props[self._name_prop]
            except KeyError:
                pass
        return self.resource_class(self, uri, name, res_props)

    @logged_api_call
    def findall(self, **filter_args):
        """
        Find zero or more resources in scope of this manager, by matching
        resource properties against the specified filter arguments, and return
        a list of their Python resource objects (e.g. for CPCs, a list of
        :class:`~zhmcclient.Cpc` objects is returned).

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way, as described
        in :meth:`~zhmcclient.BaseManager.list`.

        Authorization requirements:

        * see the `list()` method in the derived classes.

        Parameters:

          \\**filter_args:
            All keyword arguments are used as filter arguments. Specifying no
            keyword arguments causes no filtering to happen. See the examples
            for usage details.

        Returns:

          List of resource objects in scope of this manager object that match
          the filter arguments. These resource objects have a minimal set of
          properties.

        Raises:

          : Exceptions raised by the `list()` methods in derived resource
            manager classes (see :ref:`Resources`).

        Examples:

        * The following example finds partitions in a CPC by status. Because
          the 'status' resource property is also a valid Python variable name,
          there are two ways for the caller to specify the filter arguments for
          this method:

          As named parameters::

              run_states = ['active', 'degraded']
              run_parts = cpc.partitions.find(status=run_states)

          As a parameter dictionary::

              run_parts = cpc.partitions.find(**{'status': run_states})

        * The following example finds adapters of the OSA family in a CPC
          with an active status. Because the resource property for the adapter
          family is named 'adapter-family', it is not suitable as a Python
          variable name. Therefore, the caller can specify the filter argument
          only as a parameter dictionary::

              filter_args = {'adapter-family': 'osa', 'status': 'active'}
              active_osa_adapters = cpc.adapters.findall(**filter_args)
        """
        obj_list = self.list(filter_args=filter_args)
        return obj_list

    @logged_api_call
    def find(self, **filter_args):
        """
        Find exactly one resource in scope of this manager, by matching
        resource properties against the specified filter arguments, and return
        its Python resource object (e.g. for a CPC, a :class:`~zhmcclient.Cpc`
        object is returned).

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way, as described
        in :meth:`~zhmcclient.BaseManager.list`.

        Authorization requirements:

        * see the `list()` method in the derived classes.

        Parameters:

          \\**filter_args:
            All keyword arguments are used as filter arguments. Specifying no
            keyword arguments causes no filtering to happen. See the examples
            for usage details.

            If the resource name is specified in the filter arguments, it
            is matched with string comparison (i.e. not as a regular
            expression).
            The string comparison is case sensitive or case insensitive,
            dependent on the resource type.

            Any other filter arguments are ignored if the resource name is
            specified, because the name is unique within the scope of this
            resource manager.

        Returns:

          Resource object in scope of this manager object that matches the
          filter arguments. This resource object has a minimal set of
          properties.

        Raises:

          :exc:`~zhmcclient.NotFound`: No matching resource found.
          :exc:`~zhmcclient.NoUniqueMatch`: More than one matching resource
            found.
          : Exceptions raised by the `list()` methods in derived resource
            manager classes (see :ref:`Resources`).

        Examples:

        * The following example finds a CPC by its name. Because the 'name'
          resource property is also a valid Python variable name, there are
          two ways for the caller to specify the filter arguments for this
          method:

          As named parameters::

              cpc = client.cpcs.find(name='CPC001')

          As a parameter dictionary::

              filter_args = {'name': 'CPC0001'}
              cpc = client.cpcs.find(**filter_args)

        * The following example finds a CPC by its object ID. Because the
          'object-id' resource property is not a valid Python variable name,
          the caller can specify the filter argument only as a parameter
          dictionary::

              filter_args = {'object-id': '12345-abc...de-12345'}
              cpc = client.cpcs.find(**filter_args)
        """
        if self._name_prop in filter_args:
            return self.find_by_name(filter_args[self._name_prop])

        obj_list = self.findall(**filter_args)
        num_objs = len(obj_list)
        if num_objs == 0:
            raise NotFound(filter_args, self)
        if num_objs > 1:
            raise NoUniqueMatch(filter_args, self, obj_list)
        return obj_list[0]

    def list(self, full_properties=False, filter_args=None):
        """
        Find zero or more resources in scope of this manager, by matching
        resource properties against the specified filter arguments, and return
        a list of their Python resource objects (e.g. for CPCs, a list of
        :class:`~zhmcclient.Cpc` objects is returned).

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, for resources that have a List operation, the List
          operation is performed with the subset of the provided filter
          arguments that can be handled on the HMC side (this varies by
          resource type) and the remaining filter arguments are applied on the
          client side on the list result. For resources that are element objects
          without a List operation, the corresponding array property of the
          parent object is used to list the resources, and the provided filter
          arguments are applied.

        At the level of the :class:`~zhmcclient.BaseManager` class, this method
        defines the interface for the `list()` methods implemented in the
        derived resource classes.

        Authorization requirements:

        * see the `list()` method in the derived classes.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only a minimal set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments. `None` causes no filtering to happen. See the
            examples for usage details.

        Returns:

          List of resource objects in scope of this manager object that match
          the filter arguments. These resource objects have a set of properties
          according to the `full_properties` parameter.

        Raises:

          : Exceptions raised by the `list()` methods in derived resource
            manager classes (see :ref:`Resources`).

        Examples:

        * The following example finds those OSA adapters in cage '1234' of a
          given CPC, whose state is 'stand-by', 'reserved', or 'unknown'::

              filter_args = {
                  'adapter-family': 'osa',
                  'card-location': '1234-.*',
                  'state': ['stand-by', 'reserved', 'unknown'],
              }
              osa_adapters = cpc.adapters.list(full_properties=True,
                                               filter_args=filter_args)

          The returned resource objects will have the full set of properties.
        """
        raise NotImplementedError

    @logged_api_call
    def find_by_name(self, name):
        """
        Find a resource by name and return its Python resource object (e.g.
        for a CPC, a :class:`~zhmcclient.Cpc` object is returned).

        This method performs an optimized lookup that uses a name-to-URI
        mapping cached in this manager object.

        This method is automatically used by the
        :meth:`~zhmcclient.BaseManager.find` and
        :meth:`~zhmcclient.BaseManager.findall` methods, so it does not
        normally need to be used directly by users.

        Authorization requirements:

        * see the `list()` method in the derived classes.

        Parameters:

          name (string):
            Name of the resource. The name is matched with string comparison
            (i.e. not as a regular expression). The string comparison is case
            sensitive or case insensitive, dependent on the resource type.
            Must not be `None`.

        Returns:

          Resource object in scope of this manager object that has the specified
          name. This resource object has a minimal set of properties.

        Raises:

          :exc:`~zhmcclient.NotFound`: No matching resource found.
          : Exceptions raised by the `list()` methods in derived resource
            manager classes (see :ref:`Resources`).

        Examples:

        * The following example finds a CPC by its name::

              cpc = client.cpcs.find_by_name('CPC001')
        """
        name, uri = self._name_uri_cache.get(name)
        resource_props = {
            self._name_prop: name,
        }
        resource_obj = self.resource_object(uri, resource_props)
        return resource_obj

    @logged_api_call
    def find_local(self, name, uri, properties=None):
        """
        Return a local resource object without fetching it from the HMC.

        The resource object is fully functional, for example it has the proper
        parent resource object set and its properties can be fetched using
        :meth:`~zhmcclient.BaseResource.pull_full_properties`.

        If the resource object is in the name-to-URI cache, it is returned from
        there, and dependent on the history it may have a minimal set of
        properties or the full set of properties. Otherwise, a new resource
        object is constructed locally from the specified name, uri and
        properties parameters. That resource object is not put into the
        name-to-URI cache because there was no validation that the specified
        properties are up to date or even valid at all.

        Parameters:

          name (string):
            Name of the resource. The name is matched with string comparison
            (i.e. not as a regular expression). The string comparison is case
            sensitive or case insensitive, dependent on the resource type.
            Must not be `None`.

          uri (string):
            Object URI of the resource. Must not be `None`.

          properties (dict):
            Additional properties. Only used when the resource is not found
            in the name-to-URI cache. It is the responsibility of the user
            to ensure that the properties are valid.

        Returns:

          Resource object.
        """
        # Get the resource object from the name-to-URI cache, if possible.
        resource_obj = self._try_optimized_lookup(dict(name=name))
        if resource_obj:
            assert uri == resource_obj.uri
            return resource_obj

        # Create a new local resource object.
        resource_props = {
            self._name_prop: name,
        }
        resource_props.update(properties)
        resource_obj = self.resource_object(uri, resource_props)
        return resource_obj

    @logged_api_call
    def flush(self):
        """
        Invalidate the Name-URI cache of this manager.

        **Deprecated:** This method is deprecated and using it will cause a
        :exc:`~py:exceptions.DeprecationWarning` to be issued. Use
        :meth:`~zhmcclient.BaseManager.invalidate_cache` instead.
        """
        warnings.warn(
            "Use of flush() on zhmcclient manager objects is deprecated; "
            "use invalidate_cache() instead", DeprecationWarning)
        self.invalidate_cache()

    def dump(self):
        """
        Dump the resources of this resource manager as a resource definition.

        This is the default implementation for the case where the resource
        manager has no internal state that needs to be saved.
        If the resource manager does have internal state, this method needs to
        be overridden in the resource manager subclass.

        The returned resource definition of this implementation has the
        following format::

            [
                {...},  # resource definition
                ...
            ]

        Returns:

          list: Resource definitions of the resources of this resource manager.
        """
        res_list = []
        for res in self.list():
            res_list.append(res.dump())
        return res_list



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Virtual Function` is a logical entity that provides a
:term:`Partition` with access to
:term:`Accelerator Adapters <Accelerator Adapter>`.

Virtual Function resources are contained in Partition resources.

Virtual Functions only exist in :term:`CPCs <CPC>` that are in DPM mode.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_VIRTUAL_FUNCTION

__all__ = ['VirtualFunctionManager', 'VirtualFunction']


class VirtualFunctionManager(BaseManager):
    """
    Manager providing access to the
    :term:`Virtual Functions <Virtual Function>` in a particular
    :term:`Partition`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Partition` object (in DPM mode):

    * :attr:`~zhmcclient.Partition.virtual_functions`

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, partition):
        # This function should not go into the docs.
        # Parameters:
        #   partition (:class:`~zhmcclient.Partition`):
        #     Partition defining the scope for this manager.
        super().__init__(
            resource_class=VirtualFunction,
            class_name=RC_VIRTUAL_FUNCTION,
            session=partition.manager.session,
            parent=partition,
            base_uri=f'{partition.uri}/virtual-functions',
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='name',
            query_props=[],
            list_has_name=False)

    @property
    def partition(self):
        """
        :class:`~zhmcclient.Partition`: :term:`Partition` defining the scope
        for this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the Virtual Functions of this Partition.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the corresponding array property for this resource in the
          parent object is used to list the resources, and the provided filter
          arguments are applied.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.VirtualFunction` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        return self._list_with_parent_array(
            self.partition, 'virtual-function-uris', full_properties,
            filter_args)

    @logged_api_call
    def create(self, properties):
        """
        Create a Virtual Function in this Partition.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Object-access permission to the backing accelerator Adapter.
        * Task permission for the "Partition Details" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Virtual Function' in the :term:`HMC API` book.

        Returns:

          VirtualFunction:
            The resource object for the new Virtual Function.
            The object will have its 'element-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.partition.uri + '/virtual-functions',
                                   body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        vf = VirtualFunction(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return vf


class VirtualFunction(BaseResource):
    """
    Representation of a :term:`Virtual Function`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of a Virtual Function, see section
    'Data model - Virtual Function Element Object' in section
    'Partition object' in the :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.VirtualFunctionManager`).

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.VirtualFunctionManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, VirtualFunctionManager), (
            "VirtualFunction init: Expected manager type "
            f"{VirtualFunctionManager}, got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this Virtual Function.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the Partition of this Virtual Function.
        * Task permission for the "Partition Details" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self._uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))

        parent_vf_uris = self.manager.parent.get_properties_local(
            'virtual-function-uris')
        if parent_vf_uris:
            try:
                parent_vf_uris.remove(self._uri)
            except ValueError:
                pass

        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this Virtual Function.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the Partition of this Virtual Function.
        * When updating the "adapter-uri" property, object-access permission to
          the Adapter identified in that URI.
        * Task permission for the "Partition Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model - Virtual Function element object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Console` resource represents an HMC.

In a paired setup with primary and alternate HMC, each HMC is represented as
a separate :term:`Console` resource.
"""


import time

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import timestamp_from_datetime, divide_filter_args, \
    make_query_str, matches_filters, RC_CONSOLE
from ._storage_group import StorageGroupManager
from ._storage_group_template import StorageGroupTemplateManager
from ._user import UserManager
from ._user_role import UserRoleManager
from ._user_pattern import UserPatternManager
from ._password_rule import PasswordRuleManager
from ._task import TaskManager
from ._ldap_server_definition import LdapServerDefinitionManager
from ._unmanaged_cpc import UnmanagedCpcManager
from ._group import GroupManager
from ._utils import get_features
from ._certificates import CertificateManager

__all__ = ['ConsoleManager', 'Console']


class ConsoleManager(BaseManager):
    """
    Manager providing access to the :term:`Console` representing the HMC this
    client is connected to.

    In a paired setup with primary and alternate HMC, each HMC is represented
    as a separate :term:`Console` resource.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Client` object:

    * :attr:`zhmcclient.Client.consoles`

    HMC/SE version requirements: None
    """

    def __init__(self, client):
        # This function should not go into the docs.
        # Parameters:
        #   client (:class:`~zhmcclient.Client`):
        #      Client object for the HMC to be used.

        super().__init__(
            resource_class=Console,
            class_name=RC_CONSOLE,
            session=client.session,
            parent=None,
            base_uri='/api/console',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=None,
            list_has_name=False,
            supports_properties=True)
        self._client = client
        self._console = None

    @property
    def client(self):
        """
        :class:`~zhmcclient.Client`:
          The client defining the scope for this manager.
        """
        return self._client

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`:
          The :term:`Console` representing the HMC this client is connected to.

          The returned object is cached, so it is looked up only upon first
          access to this property.

          The returned object has only the following properties set:

          * 'class'
          * 'parent'
          * 'object-uri'

          Use :meth:`~zhmcclient.BaseResource.get_property` or
          :meth:`~zhmcclient.BaseResource.prop` to access any properties
          regardless of whether they are already set or first need to be
          retrieved.
        """
        if self._console is None:
            self._console = self.resource_object('/api/console')
        return self._console

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the (one) :term:`Console` representing the HMC this client is
        connected to.

        Any provided filter argument will be ignored; the `filter_args`
        parameter exists only for consistency with other list() methods.

        The listing of resources is handled by constructing a singleton
        object that represents the HMC of the current session.

        HMC/SE version requirements: None

        Authorization requirements:

        * None

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only a short set consisting of 'object-uri'.

          filter_args (dict):
            This parameter exists for consistency with other list() methods
            and will be ignored.

        Returns:

          : A list of :class:`~zhmcclient.Console` objects, containing the one
          :term:`Console` representing the HMC this client is connected to.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        uri = self._base_uri  # There is only one console object.
        if full_properties:
            props = self.session.get(uri)
        else:
            # Note: The Console resource's Object ID is not part of its URI.
            props = {
                self._uri_prop: uri,
            }
        resource_obj = self.resource_class(
            manager=self,
            uri=props[self._uri_prop],
            name=props.get(self._name_prop, None),
            properties=props)
        return [resource_obj]


class Console(BaseResource):
    """
    Representation of a :term:`Console`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.ConsoleManager`).

    HMC/SE version requirements: None
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.ConsoleManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, ConsoleManager), (
            f"Console init: Expected manager type {ConsoleManager}, got "
            f"{type(manager)}")
        super().__init__(manager, uri, name, properties)

        # The manager objects for child resources (with lazy initialization):
        self._storage_groups = None
        self._storage_group_templates = None
        self._users = None
        self._user_roles = None
        self._user_patterns = None
        self._password_rules = None
        self._tasks = None
        self._ldap_server_definitions = None
        self._unmanaged_cpcs = None
        self._groups = None
        self._certificates = None

    @property
    def storage_groups(self):
        """
        :class:`~zhmcclient.StorageGroupManager`:
          Manager object for the Storage Groups in scope of this Console.
        """
        # We do here some lazy loading.
        if not self._storage_groups:
            self._storage_groups = StorageGroupManager(self)
        return self._storage_groups

    @property
    def storage_group_templates(self):
        """
        :class:`~zhmcclient.StorageGroupTemplateManager`:
          Manager object for the Storage Group Templates in scope of this
          Console.
        """
        # We do here some lazy loading.
        if not self._storage_group_templates:
            self._storage_group_templates = StorageGroupTemplateManager(self)
        return self._storage_group_templates

    @property
    def users(self):
        """
        :class:`~zhmcclient.UserManager`: Access to the :term:`Users <User>` in
        this Console.
        """
        # We do here some lazy loading.
        if not self._users:
            self._users = UserManager(self)
        return self._users

    @property
    def user_roles(self):
        """
        :class:`~zhmcclient.UserRoleManager`: Access to the
        :term:`User Roles <User Role>` in this Console.
        """
        # We do here some lazy loading.
        if not self._user_roles:
            self._user_roles = UserRoleManager(self)
        return self._user_roles

    @property
    def user_patterns(self):
        """
        :class:`~zhmcclient.UserPatternManager`: Access to the
        :term:`User Patterns <User Pattern>` in this Console.
        """
        # We do here some lazy loading.
        if not self._user_patterns:
            self._user_patterns = UserPatternManager(self)
        return self._user_patterns

    @property
    def password_rules(self):
        """
        :class:`~zhmcclient.PasswordRuleManager`: Access to the
        :term:`Password Rules <Password Rule>` in this Console.
        """
        # We do here some lazy loading.
        if not self._password_rules:
            self._password_rules = PasswordRuleManager(self)
        return self._password_rules

    @property
    def tasks(self):
        """
        :class:`~zhmcclient.TaskManager`: Access to the :term:`Tasks <Task>` in
        this Console.
        """
        # We do here some lazy loading.
        if not self._tasks:
            self._tasks = TaskManager(self)
        return self._tasks

    @property
    def ldap_server_definitions(self):
        """
        :class:`~zhmcclient.LdapServerDefinitionManager`: Access to the
        :term:`LDAP Server Definitions <LDAP Server Definition>` in this
        Console.
        """
        # We do here some lazy loading.
        if not self._ldap_server_definitions:
            self._ldap_server_definitions = LdapServerDefinitionManager(self)
        return self._ldap_server_definitions

    @property
    def unmanaged_cpcs(self):
        """
        :class:`~zhmcclient.UnmanagedCpcManager`: Access to the unmanaged
        :term:`CPCs <CPC>` in this Console.
        """
        # We do here some lazy loading.
        if not self._unmanaged_cpcs:
            self._unmanaged_cpcs = UnmanagedCpcManager(self)
        return self._unmanaged_cpcs

    @property
    def groups(self):
        """
        :class:`~zhmcclient.GroupManager`: Access to user-defined
        :term:`Groups <Group>` in this Console.
        """
        # We do here some lazy loading.
        if not self._groups:
            self._groups = GroupManager(self)
        return self._groups

    @logged_api_call
    def restart(self, force=False, wait_for_available=True,
                operation_timeout=None):
        """
        Restart the HMC represented by this Console object.

        Once the HMC is online again, this Console object, as well as any other
        resource objects accessed through this HMC, can continue to be used.
        An automatic re-logon will be performed under the covers, because the
        HMC restart invalidates the currently used HMC session.

        HMC/SE version requirements: None

        Authorization requirements:

        * Task permission for the "Shutdown/Restart" task.
        * "Remote Restart" must be enabled on the HMC.

        Parameters:

          force (bool):
            Boolean controlling whether the restart operation is processed when
            users are connected (`True`) or not (`False`). Users in this sense
            are local or remote GUI users. HMC WS API clients do not count as
            users for this purpose.

          wait_for_available (bool):
            Boolean controlling whether this method should wait for the HMC to
            become available again after the restart, as follows:

            * If `True`, this method will wait until the HMC has restarted and
              is available again. The
              :meth:`~zhmcclient.Client.query_api_version` method will be used
              to check for availability of the HMC.

            * If `False`, this method will return immediately once the HMC
              has accepted the request to be restarted.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for HMC availability after the
            restart. The special value 0 means that no timeout is set. `None`
            means that the default async operation timeout of the session is
            used. If the timeout expires when `wait_for_available=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for the HMC to become available again after the restart.
        """
        body = {'force': force}
        self.manager.session.post(
            self.uri + '/operations/restart', resource=self, body=body)
        if wait_for_available:
            time.sleep(10)
            self.manager.client.wait_for_available(
                operation_timeout=operation_timeout)

    @logged_api_call
    def shutdown(self, force=False):
        """
        Shut down and power off the HMC represented by this Console object.

        While the HMC is powered off, any Python resource objects retrieved
        from this HMC may raise exceptions upon further use.

        In order to continue using Python resource objects retrieved from this
        HMC, the HMC needs to be started again (e.g. by powering it on
        locally). Once the HMC is available again, Python resource objects
        retrieved from that HMC can continue to be used.
        An automatic re-logon will be performed under the covers, because the
        HMC startup invalidates the currently used HMC session.

        HMC/SE version requirements:

        * HMC version >= 2.12.0

        Authorization requirements:

        * Task permission for the "Shutdown/Restart" task.
        * "Remote Shutdown" must be enabled on the HMC.

        Parameters:

          force (bool):
            Boolean controlling whether the shutdown operation is processed
            when users are connected (`True`) or not (`False`). Users in this
            sense are local or remote GUI users. HMC WS API clients do not
            count as users for this purpose.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'force': force}
        self.manager.session.post(
            self.uri + '/operations/shutdown', resource=self, body=body)

    @logged_api_call
    def make_primary(self):
        """
        Change the role of the alternate HMC represented by this Console object
        to become the primary HMC.

        If that HMC is already the primary HMC, this method does not change its
        rols and succeeds.

        The HMC represented by this Console object must participate in a
        {primary, alternate} pairing.

        HMC/SE version requirements: None

        Authorization requirements:

        * Task permission for the "Manage Alternate HMC" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        self.manager.session.post(
            self.uri + '/operations/make-primary', resource=self)

    @staticmethod
    def _time_query_parms(begin_time, end_time):
        """Return the URI query paramterer string for the specified begin time
        and end time."""
        query_parms = []
        if begin_time is not None:
            begin_ts = timestamp_from_datetime(begin_time)
            qp = f'begin-time={begin_ts}'
            query_parms.append(qp)
        if end_time is not None:
            end_ts = timestamp_from_datetime(end_time)
            qp = f'end-time={end_ts}'
            query_parms.append(qp)
        query_parms_str = '&'.join(query_parms)
        if query_parms_str:
            query_parms_str = '?' + query_parms_str
        return query_parms_str

    @logged_api_call
    def get_audit_log(self, begin_time=None, end_time=None):
        """
        Return the console audit log entries, optionally filtered by their
        creation time.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Audit and Log Management" task.

        Parameters:

          begin_time (:class:`~py:datetime.datetime`):
            Begin time for filtering. Log entries with a creation time older
            than the begin time will be omitted from the results.

            If `None`, no such filtering is performed (and the oldest available
            log entries will be included).

          end_time (:class:`~py:datetime.datetime`):
            End time for filtering. Log entries with a creation time newer
            than the end time will be omitted from the results.

            If `None`, no such filtering is performed (and the newest available
            log entries will be included).

        Returns:

          :term:`json object`:
            A JSON object with the log entries, as described in section
            'Response body contents' of operation 'Get Console Audit Log' in
            the :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms = self._time_query_parms(begin_time, end_time)
        uri = self.uri + '/operations/get-audit-log' + query_parms
        result = self.manager.session.get(uri, resource=self)
        return result

    @logged_api_call
    def get_security_log(self, begin_time=None, end_time=None):
        """
        Return the console security log entries, optionally filtered by their
        creation time.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "View Security Logs" task.

        Parameters:

          begin_time (:class:`~py:datetime.datetime`):
            Begin time for filtering. Log entries with a creation time older
            than the begin time will be omitted from the results.

            If `None`, no such filtering is performed (and the oldest available
            log entries will be included).

          end_time (:class:`~py:datetime.datetime`):
            End time for filtering. Log entries with a creation time newer
            than the end time will be omitted from the results.

            If `None`, no such filtering is performed (and the newest available
            log entries will be included).

        Returns:

          :term:`json object`:
            A JSON object with the log entries, as described in section
            'Response body contents' of operation 'Get Console Security Log' in
            the :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms = self._time_query_parms(begin_time, end_time)
        uri = self.uri + '/operations/get-security-log' + query_parms
        result = self.manager.session.get(uri, resource=self)
        return result

    @logged_api_call
    def list_unmanaged_cpcs(self, name=None):
        """
        List the unmanaged CPCs of this HMC.

        For details, see :meth:`~zhmcclient.UnmanagedCpc.list`.

        HMC/SE version requirements:

        * HMC version >= 2.13.1

        Authorization requirements: None

        Parameters:

          name (:term:`string`):
            Regular expression pattern for the CPC name, as a filter that
            narrows the list of returned CPCs to those whose name property
            matches the specified pattern.

            `None` causes no filtering to happen, i.e. all unmanaged CPCs
            discovered by the HMC are returned.

        Returns:

          : A list of :class:`~zhmcclient.UnmanagedCpc` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        filter_args = {}
        if name is not None:
            filter_args['name'] = name
        cpcs = self.unmanaged_cpcs.list(filter_args=filter_args)
        return cpcs

    @logged_api_call
    def list_permitted_partitions(
            self, full_properties=False, filter_args=None,
            additional_properties=None):
        """
        List the permitted partitions of CPCs in DPM mode managed by this HMC.

        *Added in version 1.0*

        Any CPCs in classic mode managed by the HMC will be ignored for this
        operation.

        The partitions in the result can be additionally limited by specifying
        filter arguments.

        The partitions in the result will have the following partition
        properties:

        * name (string): Name of the partition.

        * object-uri (string): Object URI of the partition.

        * type (string): Type of the partition (i.e. "linux", "ssc", "zvm").

        * status (string): Status of the partition. See the data model of
          the Partition object in the :term:`HMC API` book for values.

        * has-unacceptable-status (bool): Whether the status is unacceptable,
          according to the values in the 'acceptable-status' property.

        and the following properties from their parent CPC:

        * cpc-name (string): Name of the parent CPC of the partition.

        * cpc-object-uri (string): Object URI of the parent CPC of the
          partition.

        * se-version (string): SE version of the parent CPC of the partition,
          as M.N.U string. Note that this property is returned only on newer
          HMC 2.16 versions (HMC API version 4.10 or higher).

        HMC/SE version requirements:

        * HMC version >= 2.14.0

        Authorization requirements:

        * Object permission to the partition objects included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties for the
            returned Partition objects should be retrieved, vs. only a short
            set.

          filter_args (dict):
            Filter arguments for limiting the partitions in the result.
            `None` causes no filtering to happen.

            The following filter arguments are supported by server-side
            filtering:

            * name (string): Limits the result to partitions whose name
              match the specified regular expression.

            * type (string): Limits the result to partitions with a matching
              "type" property value (i.e. "linux", "ssc", "zvm").

            * status (string): Limits the result to partitions with a matching
              "status" property value.

            * has-unacceptable-status (bool): Limits the result to partitions
              with a matching "has-unacceptable-status" property value.

            * cpc-name (string): Limits the result to partitions whose CPC
              has a name that matches the specified regular expression.

            Any other valid property of partitions is supported by
            client-side filtering:

            * <property-name>: Any other property of partitions.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

            Using this parameter requires :ref:`API features`
            "dpm-hipersockets-partition-link-management" or
            "dpm-ctc-partition-link-management" on the HMC.

        Returns:

          : A list of :class:`~zhmcclient.Partition` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms, client_filters = divide_filter_args(
            ['name', 'type', 'status', 'has-unacceptable-status', 'cpc-name'],
            filter_args)
        if additional_properties:
            ap_parm = f"additional-properties={','.join(additional_properties)}"
            query_parms.append(ap_parm)
        query_parms_str = make_query_str(query_parms)

        # Perform the operation with the HMC, including any server-side
        # filtering.
        # Note: "List Permitted Partitions" was introduced in HMC/SE 2.14.0.
        uri = (f'{self.uri}/operations/list-permitted-partitions'
               f'{query_parms_str}')
        result = self.manager.session.get(uri, resource=self)

        cpcs_by_uri = {}  # caches local Cpc objects for CPCs already seen
        partition_objs = []
        if result:
            cpc_manager = self.manager.client.cpcs
            partition_items = result['partitions']
            for partition_item in partition_items:

                # The partition_item dicts have the following properties:
                # * name, object-uri, type, status, has-unacceptable-status,
                #   cpc-name, cpc-object-uri
                # * se-version (if HMC>=2.14.1)

                cpc_uri = partition_item['cpc-object-uri']
                try:
                    cpc = cpcs_by_uri[cpc_uri]
                except KeyError:
                    # Create a 'skeleton' local Cpc object we can hang the
                    # Partition objects off of, even if the user does not have
                    # access permissions to these CPCs. Note that different
                    # partitions can have different parent CPCs.
                    cpc_props = {}
                    cpc_props['name'] = partition_item['cpc-name']
                    if 'se-version' in partition_item:
                        cpc_props['se-version'] = partition_item['se-version']
                    cpc = cpc_manager.resource_object(cpc_uri, cpc_props)

                    cpcs_by_uri[cpc_uri] = cpc

                partition_props = dict(partition_item)
                partition_obj = cpc.partitions.resource_object(
                    partition_item['object-uri'], partition_props)

                # Apply client-side filtering
                if matches_filters(partition_obj, client_filters):
                    partition_objs.append(partition_obj)
                    if full_properties:
                        partition_obj.pull_full_properties()

        return partition_objs

    @logged_api_call
    def list_permitted_lpars(
            self, full_properties=False, filter_args=None,
            additional_properties=None):
        """
        List the permitted LPARs of CPCs in classic mode managed by this HMC.

        *Added in version 1.0*

        Any CPCs in DPM mode managed by the HMC will be ignored for this
        operation.

        The LPARs in the result can be additionally limited by specifying
        filter arguments.

        The LPARs in the result will have the following LPAR properties:

        * name (string): Name of the LPAR.

        * object-uri (string): Object URI of the LPAR.

        * activation-mode (string): Activation mode of the LPAR. See the data
          model of the Logical Partition object in the :term:`HMC API` book for
          values.

        * status (string): Status of the LPAR. See the data model of
          the Logical Partition object in the :term:`HMC API` book for values.

        * has-unacceptable-status (bool): Whether the status is unacceptable,
          according to the values in the 'acceptable-status' property.

        and the following properties from their parent CPC:

        * cpc-name (string): Name of the parent CPC of the LPAR.

        * cpc-object-uri (string): Object URI of the parent CPC of the LPAR.

        * se-version (string): SE version of the parent CPC of the LPAR, as
          M.N.U string. Note that this property is returned only on newer HMC
          2.16 versions (HMC API version 4.10 or higher).

        HMC/SE version requirements:

        * HMC version >= 2.14.0

        Authorization requirements:

        * Object permission to the LPAR objects included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties for the
            returned LPAR objects should be retrieved, vs. only a short set.

          filter_args (dict):
            Filter arguments for limiting the LPARs in the result.
            `None` causes no filtering to happen.

            The following filter arguments are supported by server-side
            filtering:

            * name (string): Limits the result to LPARs whose name
              match the specified regular expression.

            * activation-mode (string): Limits the result to LPARs with a
              matching "activation-mode" property value.

            * status (string): Limits the result to LPARs with a matching
              "status" property value.

            * has-unacceptable-status (bool): Limits the result to LPARs
              with a matching "has-unacceptable-status" property value.

            * cpc-name (string): Limits the result to LPARs whose CPC
              has a name that matches the specified regular expression.

            Any other valid property of LPARs is supported by
            client-side filtering:

            * <property-name>: Any other property of LPARs.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

            Note: This parameter is handled by the HMC starting with HMC API
            version 4.10 (HMC 2.16 GA 1.5); with older HMC API versions it is
            handled by zhmcclient.

        Returns:

          : A list of :class:`~zhmcclient.Lpar` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms, client_filters = divide_filter_args(
            ['name', 'type', 'status', 'has-unacceptable-status', 'cpc-name'],
            filter_args)

        api_version_info = self.manager.client.version_info()
        hmc_supports_additional_properties = api_version_info >= (4, 10)

        if additional_properties and hmc_supports_additional_properties:
            ap_parm = f"additional-properties={','.join(additional_properties)}"
            query_parms.append(ap_parm)
        query_parms_str = make_query_str(query_parms)

        # Perform the operation with the HMC, including any server-side
        # filtering.
        uri = (f'{self.uri}/operations/list-permitted-logical-partitions'
               f'{query_parms_str}')
        result = self.manager.session.get(uri, resource=self)

        cpcs_by_uri = {}  # caches local Cpc objects for CPCs already seen
        lpar_objs = []
        if result:
            cpc_manager = self.manager.client.cpcs
            lpar_items = result['logical-partitions']
            for lpar_item in lpar_items:

                # The partition items have the following partition properties:
                # * name, object-uri, activation-mode, status,
                #   has-unacceptable-status
                # And the following properties for their parent CPC:
                # * cpc-name (CPC property 'name')
                # * cpc-object-uri (CPC property 'object-uri')
                # * se-version (CPC property 'se-version') (if >=2.14.1)

                cpc_uri = lpar_item['cpc-object-uri']
                try:
                    cpc = cpcs_by_uri[cpc_uri]
                except KeyError:
                    # Create a 'skeleton' local Cpc object we can hang the
                    # Partition objects off of, even if the user does not have
                    # access permissions to these CPCs. Note that different
                    # partitions can have different parent CPCs.
                    cpc_props = {}
                    cpc_props['name'] = lpar_item['cpc-name']
                    if 'se-version' in lpar_item:
                        cpc_props['se-version'] = lpar_item['se-version']
                    cpc = cpc_manager.resource_object(cpc_uri, cpc_props)

                    cpcs_by_uri[cpc_uri] = cpc

                lpar_props = dict(lpar_item)
                pull_props = []
                if additional_properties:
                    for prop in additional_properties:
                        try:
                            lpar_props[prop] = lpar_item[prop]
                        except KeyError:
                            pull_props.append(prop)
                lpar_obj = cpc.lpars.resource_object(
                    lpar_item['object-uri'],
                    lpar_props,
                )
                if pull_props:
                    lpar_obj.pull_properties(pull_props)

                # Apply client-side filtering
                if matches_filters(lpar_obj, client_filters):
                    lpar_objs.append(lpar_obj)
                    if full_properties:
                        lpar_obj.pull_full_properties()

        return lpar_objs

    @logged_api_call
    def list_permitted_adapters(
            self, full_properties=False, filter_args=None,
            additional_properties=None):
        """
        List the permitted adapters of all CPCs managed by this HMC.

        The result will include all adapters of any DPM CPCs and z15 or later
        classic-mode CPCs that are managed by the targeted HMC and to which the
        user has object-access permission.

        The adapters in the result can be additionally limited by specifying
        filter arguments.

        HMC/SE version requirements:

        * HMC version >= 2.16.0

        Authorization requirements:

        * Object permission to the adapter objects included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties for the
            returned Adapter objects should be retrieved, vs. only a short
            set.

          filter_args (dict):
            Filter arguments for limiting the adapters in the result.
            `None` causes no filtering to happen.

            The following filter arguments are supported by server-side
            filtering:

            * name (string): Limits the result to adapters whose name
              match the specified regular expression.

            * adapter-id (string): Limits the result to adapters with a matching
              "adapter-id" property value (i.e. PCHID).

            * adapter-family (string): Limits the result to adapters with a
              matching "adapter-family" property value (e.g. "hipersockets").

            * type (string): Limits the result to adapters with a matching
              "type" property value (e.g. "hipersockets").

            * status (string): Limits the result to adapters with a matching
              "status" property value.

            * firmware-update-pending (bool): Limits the result to adapters with
              a matching firmware-update-pending state.

            * cpc-name (string): Limits the result to adapters whose CPC
              has a name that matches the specified regular expression.

            * dpm-enabled (bool): Limits the result to adapters whose CPC
              has a matching "dpm-enabled" property.

            Any other valid property of adapters is supported by
            client-side filtering:

            * <property-name>: Any other property of adapters.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

            Using this parameter requires :ref:`API feature <API features>`
            "adapter-network-information".

        Returns:

          : A list of :class:`~zhmcclient.Adapter` objects.

          If no additional or full properties are specified, the returned
          adapters will have the following properties:

          * object-uri, name, adapter-id, adapter-family, type, status,
          * firmware-update-pending (if CPC >=2.16 and
            LI_1580_CRYPTO_AUTO_TOGGLE feature is enabled)

          and the following properties for their parent CPC:

          * cpc-name (CPC property 'name')
          * cpc-object-uri (CPC property 'object-uri')
          * se-version (CPC property 'se-version')
          * dpm-enabled (CPC property 'dpm-enabled')

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms, client_filters = divide_filter_args(
            ['name', 'adapter-id', 'adapter-family', 'type', 'status',
             'firmware-update-pending', 'cpc-name', 'dpm-enabled'],
            filter_args)
        if additional_properties:
            ap_parm = f"additional-properties={','.join(additional_properties)}"
            query_parms.append(ap_parm)
        query_parms_str = make_query_str(query_parms)

        # Perform the operation with the HMC, including any server-side
        # filtering.
        # Note: "List Permitted Adapters" was introduced in HMC/SE 2.16.0.
        uri = (f'{self.uri}/operations/list-permitted-adapters'
               f'{query_parms_str}')
        result = self.manager.session.get(uri, resource=self)

        adapter_obj_list = []
        if result:

            # Group the returned adapters by CPC
            adapter_items_by_cpc = {}
            cpcs_by_name = {}
            for adapter_item in result['adapters']:
                cpc_name = adapter_item['cpc-name']
                if cpc_name not in adapter_items_by_cpc:
                    adapter_items_by_cpc[cpc_name] = []
                adapter_items_by_cpc[cpc_name].append(adapter_item)
                if cpc_name not in cpcs_by_name:
                    # Create a 'skeleton' local Cpc object we can hang the
                    # Adapter objects off of, even if the user does not have
                    # access permissions to these CPCs. Note that different
                    # adapters can have different parent CPCs.
                    cpc_props = {
                        'dpm-enabled': adapter_item['dpm-enabled']
                    }
                    if 'se-version' in adapter_item:
                        cpc_props['se-version'] = adapter_item['se-version']
                    cpc = self.manager.client.cpcs.find_local(
                        name=adapter_item['cpc-name'],
                        uri=adapter_item['cpc-object-uri'],
                        properties=cpc_props,
                    )
                    cpcs_by_name[cpc_name] = cpc

            # Process the returned adapters
            for cpc_name, cpc in cpcs_by_name.items():
                adapter_items = adapter_items_by_cpc[cpc_name]
                adapter_manager = cpc.adapters
                if full_properties:
                    # pylint: disable=protected-access
                    adapters = adapter_manager._get_properties_bulk(
                        adapter_items, client_filters)
                    adapter_obj_list.extend(adapters)
                else:
                    for adapter_item in adapter_items:
                        # pylint: disable=protected-access
                        adapter = adapter_manager.resource_class(
                            manager=adapter_manager,
                            uri=adapter_item[adapter_manager._uri_prop],
                            name=adapter_item[adapter_manager._name_prop],
                            properties=adapter_item)
                        if matches_filters(adapter, client_filters):
                            adapter_obj_list.append(adapter)

        return adapter_obj_list

    @logged_api_call
    def list_api_features(self, name=None):
        """
        Returns information about the :ref:`API features` available on
        this console.

        HMC/SE version requirements:

        * HMC version >= 2.16.0 with HMC API version >= 4.10

        Authorization requirements: None

        Parameters:

          name:
            A regular expression used to limit returned objects to those that
            have a matching name field.

        Returns:

          list of strings: The list of API features that are available on this
          client. For HMC API versions prior to 4.10, an empty list is returned.

        """
        # TODO: add reference to WSAPI book chapter regarding API features
        return get_features(self.manager.session, '/api/console', name)

    @property
    def certificates(self):
        """
        :class:`~zhmcclient.CertificateManager`: Access to the
        :term:`Certificates <Certificate>` in this HMC.
        """
        # We do here some lazy loading.
        if not self._certificates:
            self._certificates = CertificateManager(self)
        return self._certificates

    @logged_api_call
    def single_step_install(
            self, bundle_level=None, backup_location_type='usb',
            accept_firmware=True, ftp_host=None,
            ftp_protocol=None, ftp_user=None,
            ftp_password=None, ftp_directory=None,
            wait_for_completion=True, operation_timeout=None):
        """
        Upgrades the firmware on this HMC to a new bundle level.

        This is done by performing the "Console Single Step Install" operation
        which performs the following steps:

        * If `accept_firmware` is True, the firmware currently installed on the
          this HMC is accepted. Note that once firmware is accepted, it
          cannot be removed.
        * A backup of the this HMC is performed to the specified backup device.
        * The new firmware identified by the bundle-level field is retrieved
          from the IBM support site or from an FTP server, and installed.
        * The newly installed firmware is activated, which includes rebooting
          this HMC.

        Note that it is not possible to downgrade the HMC firmware with this
        operation.

        For HMCs that run on an HMA that also hosts an SE (e.g. z16 and higher),
        the HMC firmware can only be upgraded if the HMA hosts an alternate SE.

        HMC/SE version requirements:

        * HMC version >= 2.16.0

        Authorization requirements:

        * Task permission to the "Single Step Console Internal Code" task.

        Parameters:

          bundle_level (string): Name of the bundle to be installed on the HMC
            (e.g. 'H71').
            If `None`, all locally available code changes, or in case of
            retrieving code changes from an FTP server, all code changes on
            the FTP server, will be installed.

          backup_location_type (string): Type of backup location for the
            HMC backup that is performed:

              - "ftp": The FTP server that was used for the last console backup
                as defined on the "Configure Backup Settings" user interface
                task in the HMC GUI.
              - "usb": The USB storage device mounted to the HMC.

          accept_firmware (bool): Accept the previous bundle level before
            installing the new level.

          ftp_host (string): The hostname for the FTP server from which
            the firmware will be retrieved, or `None` to retrieve it from the
            IBM support site.

          ftp_protocol (string): The protocol to connect to the FTP
            server, if the firmware will be retrieved from an FTP server,
            or `None`. Valid values are: "ftp", "ftps", "sftp".

          ftp_user (string): The username for the FTP server login,
            if the firmware will be retrieved from an FTP server, or `None`.

          ftp_password (string): The password for the FTP server login,
            if the firmware will be retrieved from an FTP server, or `None`.

          ftp_directory (string): The path name of the directory on the
            FTP server with the firmware files,
            if the firmware will be retrieved from an FTP server, or `None`.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation including any HMC
            restarts, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation including any HMC
              restarts.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation including any HMC restarts. The
            special value 0 means that no timeout is set. `None` means that
            the default async operation timeout of the session is used. If the
            timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC. The Job object will be valid across
            any HMC restarts that occur during the upgrade operation.

            This job supports cancellation. Note there are only a few
            interruption points in the firmware install process, so it may be
            some time before the job is canceled, and after some point, will
            continue on to completion. The job status and reason codes will
            indicate whether the job was canceled or ran to completion. If the
            job is successfully canceled, any steps that were successfully
            completed will not be rolled back.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """

        # The 'wait_for_completion' parameter from 2.12.0 became
        # 'ftp_host' after that, so we detect the passing of
        # 'wait_for_completion' as a positional argument.
        assert ftp_host is None or \
            isinstance(ftp_host, str)

        body = {
            'backup-location-type': backup_location_type,
            'accept-firmware': accept_firmware,
        }

        if bundle_level is not None:
            body['bundle-level'] = bundle_level

        if ftp_host is not None:
            body['ftp-retrieve'] = True
            body['ftp-server-host'] = ftp_host
            body['ftp-server-user'] = ftp_user
            body['ftp-server-password'] = ftp_password
            body['ftp-server-directory'] = ftp_directory
            body['ftp-server-protocol'] = ftp_protocol

        result = self.manager.session.post(
            self.uri + '/operations/single-step-install', resource=self,
            body=body, wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)

        return result

    @logged_api_call
    def delete_uninstalled_firmware(
            self, ec_levels=None,
            wait_for_completion=True, operation_timeout=None):
        """
        Deletes retrieved but uninstalled firmware on this HMC.

        This is done by performing the "Console Delete Retrieved Internal Code"
        operation.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "hmc-delete-retrieved-internal-code"

        Authorization requirements:

        * Task permission to the "Change Console Internal Code" task.

        Parameters:

          ec_levels (list of tuple(ec,mcl)): The specific EC levels back to
            which the firmware should be deleted.
            If `None`, all uninstalled MCLs are deleted.

            The EC levels are specified as a list of tuples (ec, mcl) where:

              - ec (string): EC number of the EC stream (e.g. "P30719")
              - mcl (string): MCL number within the EC stream (e.g. "001")

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation including any HMC
            restarts, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation including any HMC
              restarts.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation including any HMC restarts. The
            special value 0 means that no timeout is set. `None` means that
            the default async operation timeout of the session is used. If the
            timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          string or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns a string with the
            message text describing the detailed error that occurred when the
            operation was not successful.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC. The Job object will be valid across
            any HMC restarts that occur during the upgrade operation.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """

        body = {}

        if ec_levels is not None:
            body['ec-levels'] = \
                [{"number": ec[0], "mcl": ec[1]} for ec in ec_levels]

        result = self.manager.session.post(
            self.uri + '/operations/delete-retrieved-internal-code',
            resource=self,
            body=body, wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)

        return result

    def dump(self):
        """
        Dump this Console resource with its properties and child resources
        (recursively) as a resource definition.

        The returned resource definition has the following format::

            {
                # Resource properties:
                "properties": {...},

                # Child resources:
                "users": [...],
                "user_roles": [...],
                "user_patterns": [...],
                "password_rules": [...],
                "tasks": [...],
                "ldap_server_definitions": [...],
                "unmanaged_cpcs": [...],
                "storage_groups": [...],
            }

        Returns:

          dict: Resource definition of this resource.
        """

        # Dump the resource properties
        resource_dict = super().dump()

        # Dump the child resources
        users = self.users.dump()
        if users:
            resource_dict['users'] = users
        user_roles = self.user_roles.dump()
        if user_roles:
            resource_dict['user_roles'] = user_roles
        user_patterns = self.user_patterns.dump()
        if user_patterns:
            resource_dict['user_patterns'] = user_patterns
        password_rules = self.password_rules.dump()
        if password_rules:
            resource_dict['password_rules'] = password_rules
        tasks = self.tasks.dump()
        if tasks:
            resource_dict['tasks'] = tasks
        ldap_server_definitions = self.ldap_server_definitions.dump()
        if ldap_server_definitions:
            resource_dict['ldap_server_definitions'] = ldap_server_definitions
        storage_groups = self.storage_groups.dump()
        if storage_groups:
            resource_dict['storage_groups'] = storage_groups

        # Note: Unmanaged CPCs are not dumped, since their properties cannot
        #       be retrieved.

        return resource_dict



#!/usr/bin/env python
# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that cleans up zhmc test partitions left over by the other examples.
"""

import sys
import re
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    print("Deleting leftover zhmc test partitions created by the other "
          "examples ...")
    try:
        parts = client.consoles.console.list_permitted_partitions()
        for part in parts:
            if re.match(r'^zhmc_test_[a-z0-9\-]{8,}$', part.name):
                if part.get_property('status') != 'stopped':
                    print(f"Stopping test partition: {part.name} ...")
                    part.stop()
                print(f"Deleting test partition: {part.name} ...")
                part.delete()
    except zhmcclient.Error as exc:
        print(f"Error: {exc.__class__.__name__}: {exc}")
        sys.exit(1)

finally:
    print("Logging off ...")
    session.logoff()



#!/usr/bin/env python
# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Support for the :ref:`auto-updating` of Python zhmcclient resource and manager
objects based on HMC notifications.
"""


import logging
import json
from json import JSONDecodeError
import ssl

from ._constants import DEFAULT_STOMP_PORT, JMS_LOGGER_NAME, \
    DEFAULT_STOMP_CONNECT_TIMEOUT, \
    DEFAULT_STOMP_CONNECT_RETRIES, DEFAULT_STOMP_RECONNECT_SLEEP_INITIAL, \
    DEFAULT_STOMP_RECONNECT_SLEEP_INCREASE, DEFAULT_STOMP_RECONNECT_SLEEP_MAX, \
    DEFAULT_STOMP_RECONNECT_SLEEP_JITTER, DEFAULT_STOMP_KEEPALIVE, \
    DEFAULT_STOMP_HEARTBEAT_SEND_CYCLE, DEFAULT_STOMP_HEARTBEAT_RECEIVE_CYCLE, \
    DEFAULT_STOMP_HEARTBEAT_RECEIVE_CHECK
from ._utils import RC_CPC, RC_CHILDREN_CLIENT, RC_CHILDREN_CPC, \
    RC_CHILDREN_CONSOLE, get_stomp_rt_kwargs, get_headers_message
from ._client import Client
from ._manager import BaseManager
from ._resource import BaseResource
from ._notification import StompRetryTimeoutConfig

__all__ = ['AutoUpdater']

JMS_LOGGER = logging.getLogger(JMS_LOGGER_NAME)


class AutoUpdater:
    """
    A class that automatically updates

    * the properties of zhmcclient resource objects that are enabled for
      auto-updating
    * the list of zhmcclient resource objects in manager objects that are
      enabled for auto-updating

    based on respective notifications from the HMC.

    **Experimental:** This class is considered experimental at this point, and
    its API may change incompatibly as long as it is experimental.

    Note: The user should not create any objects of this class nor invoke any
    methods of this class, because the objects are created automatically
    when a :class:`~zhmcclient.Session` object is subscribed for
    auto-updating via its :meth:`~zhmcclient.Session.subscribe_auto_update`
    method.

    Creating an object of this class performs a logon to the HMC to retrieve
    the notification topics, and then establishes a JMS session with the HMC and
    subscribes for the object notification topic of the session. This causes
    the HMC to emit status notifications, property notifications, and inventory
    notifications, which are processed by this class and cause the updates
    to happen.

    Resource objects can be enabled for auto-updating via their
    :meth:`~zhmcclient.BaseResource.enable_auto_update` method.

    Manager objects can be enabled for auto-updating via their
    :meth:`~zhmcclient.BaseManager.enable_auto_update` method.

    Zhmcclient resource objects or manager objects that are not enabled for
    auto-updating remain unchanged.

    HMC/SE version requirements: None
    """

    default_stomp_rt_config = StompRetryTimeoutConfig(
        connect_timeout=DEFAULT_STOMP_CONNECT_TIMEOUT,
        connect_retries=DEFAULT_STOMP_CONNECT_RETRIES,
        reconnect_sleep_initial=DEFAULT_STOMP_RECONNECT_SLEEP_INITIAL,
        reconnect_sleep_increase=DEFAULT_STOMP_RECONNECT_SLEEP_INCREASE,
        reconnect_sleep_max=DEFAULT_STOMP_RECONNECT_SLEEP_MAX,
        reconnect_sleep_jitter=DEFAULT_STOMP_RECONNECT_SLEEP_JITTER,
        keepalive=DEFAULT_STOMP_KEEPALIVE,
        heartbeat_send_cycle=DEFAULT_STOMP_HEARTBEAT_SEND_CYCLE,
        heartbeat_receive_cycle=DEFAULT_STOMP_HEARTBEAT_RECEIVE_CYCLE,
        heartbeat_receive_check=DEFAULT_STOMP_HEARTBEAT_RECEIVE_CHECK,
    )

    def __init__(self, session, stomp_rt_config=None):
        """
        Parameters:

          session (:class:`~zhmcclient.Session`): Session for which the
            auto updater should do its work. This defines the HMC host
            and credentials that are used to establish the JMS session with
            the HMC. The session may or may not be logged on.

          stomp_rt_config (:class:`~zhmcclient.StompRetryTimeoutConfig`):
            The STOMP retry/timeout configuration for this session, overriding
            any defaults.

            `None` for an attribute in that configuration object means that the
            default value will be used for that attribute.

            `None` for the entire `stomp_rt_config` parameter means that a
            default configuration will be used with the default values for all
            of its attributes.

            See :ref:`Constants` for the default values.
        """

        self._session = session
        self._rt_config = stomp_rt_config

        # STOMP connection
        self._conn = None

        # Registered resource and manager objects, as:
        #   dict(key: uri, value: dict(key: id, value: object))
        self._registered_objects = {}

        # Subscription ID. We use some value that allows to identify on the
        # HMC that this is the zhmcclient, but otherwise we are not using
        # this value ourselves.
        self._sub_id = f'zhmcclient.{id(self)}'

        # Lazy importing of the stomp module, because the import is slow in some
        # versions.
        # pylint: disable=import-outside-toplevel
        import stomp
        self._stomp = stomp

    def open(self):
        """
        Open the JMS session with the HMC.

        This creates a STOMP connection with the actual HMC of the session and
        subscribes to the object notification topic.

        If the session does not yet have an object notification topic set,
        the session is logged on.
        """
        if not self._session.object_topic:
            self._session.logon()  # This sets actual_host

        rt_kwargs = get_stomp_rt_kwargs(self._rt_config)
        self._conn = self._stomp.Connection(
            [(self._session.actual_host, DEFAULT_STOMP_PORT)], **rt_kwargs)
        set_kwargs = dict()
        set_kwargs['ssl_version'] = ssl.PROTOCOL_TLS_CLIENT
        self._conn.set_ssl(
            for_hosts=[(self._session.actual_host, DEFAULT_STOMP_PORT)],
            **set_kwargs)

        listener = _UpdateListener(self, self._session)
        self._conn.set_listener('', listener)
        # pylint: disable=protected-access
        self._conn.connect(self._session.userid, self._session._password,
                           wait=True)

        dest = "/topic/" + self._session.object_topic
        self._conn.subscribe(destination=dest, id=self._sub_id, ack='auto')

        listener.init_cpcs()

        JMS_LOGGER.info(
            "JMS session for object notification topic '%s' has been "
            "established", self._session.object_topic)

    def close(self):
        """
        Close the JMS session with the HMC.

        This implicitly unsubscribes from the object notification topic this
        auto updater was created for.
        """
        self._conn.disconnect()
        self._conn = None

        JMS_LOGGER.info(
            "JMS session for object notification topic '%s' has been "
            "disconnected", self._session.object_topic)

    def is_open(self):
        """
        Return whether the JMS session with the HMC is open.
        """
        return self._conn is not None

    def register_object(self, obj):
        """
        Register a resource or manager object to this auto updater.

        If this object (identified by its Python id) is already registered,
        nothing is done.
        """
        assert isinstance(obj, (BaseResource, BaseManager))
        uri = obj.uri
        res_id = id(obj)
        if uri not in self._registered_objects:
            self._registered_objects[uri] = {}
        id_dict = self._registered_objects[uri]
        if res_id not in id_dict:
            id_dict[res_id] = obj

    def unregister_object(self, obj):
        """
        Unregister a resource or manager object from this auto updater.

        If this object (identified by its Python id) is already unregistered,
        nothing is done.
        """
        assert isinstance(obj, (BaseResource, BaseManager))
        uri = obj.uri
        res_id = id(obj)
        if uri in self._registered_objects:
            id_dict = self._registered_objects[uri]
            if res_id in id_dict:
                del id_dict[res_id]
            if not id_dict:
                del self._registered_objects[uri]

    def registered_objects(self, uri):
        """
        Generator that yields the resource or manager objects for the specified
        URI.
        """
        if uri in self._registered_objects:
            id_dict = self._registered_objects[uri]
            # pylint: disable=use-yield-from
            yield from id_dict.values()

    def has_objects(self):
        """
        Return boolean indicating whether there are any resource objects
        registered.
        """
        return bool(self._registered_objects)


class _UpdateListener:
    # pylint: disable=too-few-public-methods
    """
    A notification listener class for use by the Python `stomp` package.

    This is an internal class that does not need to be accessed or created by
    the user. An object of this class is automatically created by the
    :class:`~zhmcclient.AutoUpdater` class, for its notification
    topic.

    Note: In the stomp examples, this class inherits from
    stomp.ConnectionListener. However, since that class defines only empty
    methods and since we want to import the stomp module in a lazy manner,
    we are not using that class, and stomp does not require us to.
    """

    def __init__(self, updater, session):
        self._updater = updater
        self._session = session
        self._client = None

        # Lazy importing of the stomp module, because the import is slow in some
        # versions.
        # pylint: disable=import-outside-toplevel
        import stomp
        self._stomp = stomp

    def init_cpcs(self):
        """
        Initialize the CPC manager, for later use when receiving inventory
        notifications for child objects of CPCs.
        """
        if self._client is None:
            self._client = Client(self._session)
            self._client.cpcs.enable_auto_update()

    def _manager_uri_from_notification(self, headers):
        """
        Return the manager URI from the headers of an inventory notification
        and the resource URI.
        """

        try:
            res_class = headers['class']
        except KeyError:
            JMS_LOGGER.error(
                "JMS message for object notification topic '%s' "
                "has no 'class' field in "
                "its headers (ignored): %r",
                self._session.object_topic, headers)
            return None

        if 'element-uri' in headers:
            # The notification is about an element resource. In that case,
            # 'object-uri' identifies the containing (=parent) object resource.

            try:
                parent_uris = [headers['object-uri']]
            except KeyError:
                JMS_LOGGER.error(
                    "JMS message for object notification topic '%s' "
                    "has an 'element-uri' field but no 'object-uri' field in "
                    "its headers (ignored): %r",
                    self._session.object_topic, headers)
                return None

        else:
            # The notification is about an object (non-element) resource.
            # The parent object's URI is not specified in the notification
            # in this case, but there are not too many parent resources that
            # have object resources as children, so we try to find it.

            # In the removal case, it is not important what the parent
            # resource is. We simply return all possible parent resources
            # and the caller will look them all up to find the resource to
            # be removed.
            # In the addition case, it is important to find the parent object
            # because only the parent object has the right manager object into
            # which the new resource is added.

            if res_class == RC_CPC:
                # Could be a managed or unmanaged CPC
                parent_uris = ['/api/console', '/']
            elif res_class in RC_CHILDREN_CLIENT:
                # RC_CHILDREN_CLIENT includes RC_CPC, but that is already
                # processed
                parent_uris = ['/']
            elif res_class in RC_CHILDREN_CPC:
                cpcs = self._client.cpcs.list_resources_local()
                parent_uris = [cpc.uri for cpc in cpcs]
            elif res_class in RC_CHILDREN_CONSOLE:
                # RC_CHILDREN_CONSOLE includes RC_CPC, but that is already
                # processed
                parent_uris = ['/api/console']
            else:
                JMS_LOGGER.error(
                    "JMS message for object notification topic '%s' "
                    "has an unknown 'class' field in "
                    "its headers (ignored): %r",
                    self._session.object_topic, headers)
                return None

        mgr_uris = [f'{_uri}#{res_class}' for _uri in parent_uris]
        return mgr_uris

    def _get_uri(self, headers):
        """
        Return the uri of the object or element resource o which the
        notification applies.

        If the notification is about an element resource, element-uri is
        present and is the URI of that resource, and object-uri is the URI of
        the parent (containing) object.

        If the notification is about an object (=non-element) resource,
        element-uri is not present, and object-uri is the URI of that
        resource.
        """
        if 'element-uri' in headers:
            # The notification is about an element resource
            uri = headers['element-uri']
        else:
            # The notification is about an object (non-element) resource
            try:
                uri = headers['object-uri']
            except KeyError:
                JMS_LOGGER.error(
                    "JMS message for object notification topic '%s' "
                    "has no 'element-uri' field and no 'object-uri' field in "
                    "its headers (ignored): %r",
                    self._session.object_topic, headers)
                return None
        return uri

    def on_message(self, *frame_args):
        """
        Event method that gets called when this listener has received a JMS
        message (representing an HMC notification).

        Parameters:

          frame_args: The STOMP frame. For details, see get_headers_message().
        """
        headers, message = get_headers_message(frame_args)

        noti_type = headers['notification-type']
        if noti_type == 'property-change':
            try:
                msg_obj = json.loads(message)
            except JSONDecodeError:
                JMS_LOGGER.error(
                    "JMS message for object notification topic '%s' "
                    "has a non-JSON message body (ignored): %r",
                    self._session.object_topic, message)
                return
            uri = self._get_uri(headers)
            if uri is None:
                # Some error - details are already logged
                return
            JMS_LOGGER.debug(
                "JMS message for property change notification for topic '%s' "
                "for resource %s with change reports: %r",
                self._session.object_topic, uri, msg_obj['change-reports'])
            # Build the latest values from all change records. They are ordered
            # old to new.
            new_props = {}
            for cr in msg_obj['change-reports']:
                new_props[cr['property-name']] = cr['new-value']
            for obj in self._updater.registered_objects(uri):
                if obj.auto_update_enabled():
                    obj.update_properties_local(new_props)
        elif noti_type == 'status-change':
            try:
                msg_obj = json.loads(message)
            except JSONDecodeError:
                JMS_LOGGER.error(
                    "JMS message for object notification topic '%s' "
                    "has a non-JSON message body (ignored): %r",
                    self._session.object_topic, message)
                return
            uri = self._get_uri(headers)
            if uri is None:
                # Some error - details are already logged
                return
            JMS_LOGGER.debug(
                "JMS message for status change notification for topic '%s' "
                "for resource %s with change reports: %r",
                self._session.object_topic, uri, msg_obj['change-reports'])
            # Build the latest values from all change records. They are ordered
            # old to new.
            new_props = {}
            for cr in msg_obj['change-reports']:
                if 'new-status' in cr:
                    new_props['status'] = cr['new-status']
                if 'new-additional-status' in cr:
                    new_props['additional-status'] = cr['new-additional-status']
                if 'has-unacceptable-status' in cr:
                    new_props['has-unacceptable-status'] = \
                        cr['has-unacceptable-status']
            for obj in self._updater.registered_objects(uri):
                if obj.auto_update_enabled():
                    obj.update_properties_local(new_props)
        elif noti_type == 'inventory-change':
            uri = self._get_uri(headers)
            if uri is None:
                # Some error - details are already logged
                return
            action = headers['action']
            JMS_LOGGER.debug(
                "JMS message for inventory change notification for topic '%s' "
                "for resource %s with action: %r",
                self._session.object_topic, uri, action)
            if action == 'add':
                mgr_uris = self._manager_uri_from_notification(headers)
                if mgr_uris is None:
                    # Some error - details are already logged
                    return
                for mgr_uri in mgr_uris:
                    for mgr_obj in self._updater.registered_objects(mgr_uri):
                        if mgr_obj.auto_update_enabled():
                            mgr_obj.auto_update_trigger_pull()
            elif action == 'remove':
                mgr_uris = self._manager_uri_from_notification(headers)
                if mgr_uris is None:
                    # Some error - details are already logged
                    return
                for mgr_uri in mgr_uris:
                    for mgr_obj in self._updater.registered_objects(mgr_uri):
                        if mgr_obj.auto_update_enabled():
                            mgr_obj.remove_resource_local(uri)
                for obj in self._updater.registered_objects(uri):
                    if obj.auto_update_enabled():
                        obj.cease_existence_local()
            else:
                JMS_LOGGER.error(
                    "JMS message for inventory change notification specifies "
                    "unknown action %r (ignored)",
                    action)
        else:
            JMS_LOGGER.warning(
                "JMS message for notification of type %s for topic '%s' "
                "is ignored",
                noti_type, self._session.object_topic)

    def on_error(self, *frame_args):
        """
        Event method that gets called when this listener has received a JMS
        error. This happens for example when the client registers for a
        non-existing topic.

        Parameters:

          frame_args: The STOMP frame. For details, see get_headers_message().
        """
        _, message = get_headers_message(frame_args)
        JMS_LOGGER.error(
            "JMS error message received for object notification topic '%s' "
            "(ignored): %s",
            self._session.object_topic, message)

    def on_disconnected(self):
        """
        Event method that gets called when the JMS session has been
        disconnected.
        """
        JMS_LOGGER.info(
            "JMS session for object notification topic '%s' has been "
            "disconnected",
            self._session.object_topic)



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _activation_profile module.
"""


import copy
import re
import logging
import pytest

from zhmcclient import Client, ActivationProfile
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources, assert_blanked_in_message


class TestActivationProfile:
    """
    All tests for the ActivationProfile and ActivationProfileManager classes.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked CPC in classic mode,
        and add two faked activation profiles of each type.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_cpc = self.session.hmc.cpcs.add({
            'object-id': 'fake-cpc1-oid',
            # object-uri is set up automatically
            'parent': None,
            'class': 'cpc',
            'name': 'fake-cpc1-name',
            'description': 'CPC #1 (classic mode)',
            'status': 'active',
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'iml-mode': 'lpar',
        })
        self.cpc = self.client.cpcs.find(name='fake-cpc1-name')

        self.faked_reset_ap_1 = self.faked_cpc.reset_activation_profiles.add({
            # element-uri is set up automatically
            'name': 'rap_1',
            'parent': self.faked_cpc.uri,
            'class': 'reset-activation-profile',
            'description': 'RAP #1',
        })
        self.faked_reset_ap_2 = self.faked_cpc.reset_activation_profiles.add({
            # element-uri is set up automatically
            'name': 'rap_2',
            'parent': self.faked_cpc.uri,
            'class': 'reset-activation-profile',
            'description': 'RAP #2',
        })
        self.faked_image_ap_1 = self.faked_cpc.image_activation_profiles.add({
            # element-uri is set up automatically
            'name': 'iap_1',
            'parent': self.faked_cpc.uri,
            'class': 'image-activation-profile',
            'description': 'IAP #1',
            'ipl-address': '0010',
        })
        self.faked_image_ap_2 = self.faked_cpc.image_activation_profiles.add({
            # element-uri is set up automatically
            'name': 'iap_2',
            'parent': self.faked_cpc.uri,
            'class': 'image-activation-profile',
            'description': 'IAP #2',
            'ipl-address': '0010',
        })
        self.faked_load_ap_1 = self.faked_cpc.load_activation_profiles.add({
            # element-uri is set up automatically
            'name': 'lap_1',
            'parent': self.faked_cpc.uri,
            'class': 'load-activation-profile',
            'description': 'LAP #1',
        })
        self.faked_load_ap_2 = self.faked_cpc.load_activation_profiles.add({
            # element-uri is set up automatically
            'name': 'lap_2',
            'parent': self.faked_cpc.uri,
            'class': 'load-activation-profile',
            'description': 'LAP #2',
        })

    @pytest.mark.parametrize(
        "profile_type", ['reset', 'image', 'load']
    )
    def test_profilemanager_initial_attrs(self, profile_type):
        """Test initial attributes of ActivationProfileManager."""

        mgr_attr = profile_type + '_activation_profiles'
        profile_mgr = getattr(self.cpc, mgr_attr)

        # Verify all public properties of the manager object
        assert profile_mgr.resource_class == ActivationProfile
        assert profile_mgr.session == self.session
        assert profile_mgr.parent == self.cpc
        assert profile_mgr.cpc == self.cpc
        assert profile_mgr.profile_type == profile_type

    # TODO: Test for ActivationProfileManager.__repr__()

    @pytest.mark.parametrize(
        "profile_type", ['reset', 'image', 'load']
    )
    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            ({},
             ['name', 'element-uri']),
            (dict(full_properties=False),
             ['name', 'element-uri']),
            (dict(full_properties=True),
             None),
        ]
    )
    def test_profilemanager_list_full_properties(
            self, full_properties_kwargs, prop_names, profile_type):
        """Test ActivationProfileManager.list() with full_properties."""

        mgr_attr = profile_type + '_activation_profiles'
        faked_profile_mgr = getattr(self.faked_cpc, mgr_attr)
        exp_faked_profiles = faked_profile_mgr.list()
        profile_mgr = getattr(self.cpc, mgr_attr)

        # Execute the code to be tested
        profiles = profile_mgr.list(**full_properties_kwargs)

        assert_resources(profiles, exp_faked_profiles, prop_names)

    @pytest.mark.parametrize(
        "profile_type, filter_args, exp_names", [
            ('reset',
             {'name': 'rap_2'},
             ['rap_2']),
            ('reset',
             {'name': '.*rap_1'},
             ['rap_1']),
            ('reset',
             {'name': 'rap_1.*'},
             ['rap_1']),
            ('reset',
             {'name': 'rap_.'},
             ['rap_1', 'rap_2']),
            ('reset',
             {'name': '.ap_1'},
             ['rap_1']),
            ('reset',
             {'name': '.+'},
             ['rap_1', 'rap_2']),
            ('reset',
             {'name': 'rap_1.+'},
             []),
            ('reset',
             {'name': '.+rap_1'},
             []),
            ('image',
             {'name': 'iap_1'},
             ['iap_1']),
            ('image',
             {'name': '.*iap_1'},
             ['iap_1']),
            ('image',
             {'name': 'iap_1.*'},
             ['iap_1']),
            ('image',
             {'name': 'iap_.'},
             ['iap_1', 'iap_2']),
            ('image',
             {'name': '.ap_1'},
             ['iap_1']),
            ('image',
             {'name': '.+'},
             ['iap_1', 'iap_2']),
            ('image',
             {'name': 'iap_1.+'},
             []),
            ('image',
             {'name': '.+iap_1'},
             []),
            ('load',
             {'name': 'lap_2'},
             ['lap_2']),
            ('load',
             {'name': '.*lap_1'},
             ['lap_1']),
            ('load',
             {'name': 'lap_1.*'},
             ['lap_1']),
            ('load',
             {'name': 'lap_.'},
             ['lap_1', 'lap_2']),
            ('load',
             {'name': '.ap_1'},
             ['lap_1']),
            ('load',
             {'name': '.+'},
             ['lap_1', 'lap_2']),
            ('load',
             {'name': 'lap_1.+'},
             []),
            ('load',
             {'name': '.+lap_1'},
             []),
            ('reset',
             {'class': 'reset-activation-profile'},
             ['rap_1', 'rap_2']),
            ('image',
             {'class': 'image-activation-profile'},
             ['iap_1', 'iap_2']),
            ('load',
             {'class': 'load-activation-profile'},
             ['lap_1', 'lap_2']),
            ('reset',
             {'class': 'reset-activation-profile',
              'description': 'RAP #2'},
             ['rap_2']),
            ('image',
             {'class': 'image-activation-profile',
              'description': 'IAP #1'},
             ['iap_1']),
            ('load',
             {'class': 'load-activation-profile',
              'description': 'LAP #2'},
             ['lap_2']),
            ('reset',
             {'description': 'RAP #1'},
             ['rap_1']),
            ('image',
             {'description': 'IAP #2'},
             ['iap_2']),
            ('load',
             {'description': 'LAP #1'},
             ['lap_1']),
        ]
    )
    def test_profilemanager_list_filter_args(
            self, profile_type, filter_args, exp_names):
        """Test ActivationProfileManager.list() with filter_args."""

        mgr_attr = profile_type + '_activation_profiles'
        profile_mgr = getattr(self.cpc, mgr_attr)

        # Execute the code to be tested
        profiles = profile_mgr.list(filter_args=filter_args)

        assert len(profiles) == len(exp_names)
        if exp_names:
            names = [ap.properties['name'] for ap in profiles]
            assert set(names) == set(exp_names)

    @pytest.mark.parametrize(
        "list_kwargs, prop_names", [
            ({},
             ['element-uri', 'name']),
            (dict(additional_properties=[]),
             ['element-uri', 'name']),
            (dict(additional_properties=['description']),
             ['element-uri', 'name', 'description']),
            (dict(additional_properties=['description', 'ipl-address']),
             ['element-uri', 'name', 'description', 'ipl-address']),
            (dict(additional_properties=['ssc-host-name']),
             ['element-uri', 'name', 'ssc-host-name']
             # ssc-host-name is not on every image profile
             ),
        ]
    )
    def test_profilemanager_list_add_props(
            self, list_kwargs, prop_names):
        """
        Test ActivationProfileManager.list() for image profiles with
        additional_properties.
        """
        mgr_attr = 'image_activation_profiles'
        profile_mgr = getattr(self.cpc, mgr_attr)

        # Execute the code to be tested
        profiles = profile_mgr.list(**list_kwargs)

        exp_faked_profiles = [self.faked_image_ap_1, self.faked_image_ap_2]

        assert_resources(profiles, exp_faked_profiles, prop_names)

    # TODO: Test for initial ActivationProfile attributes

    def test_profile_repr(self):
        """Test ActivationProfile.__repr__()."""

        # We test __repr__() just for reset activation profiles, because the
        # ActivationProfile class is the same for all profile types and we know
        # that __repr__() does not depend on the profile type.
        profile_mgr = self.cpc.reset_activation_profiles
        reset_ap = profile_mgr.find(name='rap_1')

        # Execute the code to be tested
        repr_str = repr(reset_ap)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{reset_ap.__class__.__name__}\s+at\s+'
            rf'0x{id(reset_ap):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "profile_type", ['reset', 'image', 'load']
    )
    @pytest.mark.parametrize(
        "input_props", [
            {},
            {'description': 'New profile description'},
            {'description': ''},
            {'ssc-network-info': {
                'chpid': '1a',
                'port': 0,
                'ipaddr-type': 'dhcp',
                'vlan-id': None,
                'static-ip-info': None,
            }},
            {'group-profile-uri': None},
            {'zaware-gateway-info': None},
            {'ssc-master-pw': 'bla', 'zaware-master-pw': 'bla'},
        ]
    )
    def test_profile_update_properties(self, caplog, input_props, profile_type):
        """Test ActivationProfile.update_properties()."""

        logger_name = "zhmcclient.api"
        caplog.set_level(logging.DEBUG, logger=logger_name)

        mgr_attr = profile_type + '_activation_profiles'
        profile_mgr = getattr(self.cpc, mgr_attr)

        profile = profile_mgr.list()[0]

        profile.pull_full_properties()
        saved_properties = copy.deepcopy(profile.properties)

        # Execute the code to be tested
        profile.update_properties(properties=input_props)

        # Get its API call log record
        call_record = caplog.records[-2]

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in profile.properties
            prop_value = profile.properties[prop_name]
            assert prop_value == exp_prop_value

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        profile.pull_full_properties()
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in profile.properties
            prop_value = profile.properties[prop_name]
            assert prop_value == exp_prop_value

        # Verify the API call log record for blanked-out properties.
        assert_blanked_in_message(
            call_record.message, input_props,
            ['ssc-master-pw', 'zaware-master-pw'])



# Copyright 2019,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for testing whether the HMCs in the HMC inventory file match
reality.
"""


from requests.packages import urllib3
import pytest

import zhmcclient
# pylint: disable=unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401
from zhmcclient.testutils import HMCDefinitions

from .utils import is_cpc_property_hmc_inventory

urllib3.disable_warnings()


def test_hmcdef_cpcs(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test that the HMC actually manages the CPCs defined in its inventory file
    and that these CPCs actually have the properties defined there.
    """
    client = zhmcclient.Client(hmc_session)
    hd = hmc_session.hmc_definition

    cpcs = client.cpcs.list()
    cpc_names = [cpc.name for cpc in cpcs]

    for cpc_name in hd.cpcs:
        def_cpc_props = dict(hd.cpcs[cpc_name])

        print(f"Checking CPC {cpc_name}")

        assert cpc_name in cpc_names, (
            f"CPC {cpc_name} defined in inventory file for HMC {hd.nickname!r} "
            f"at {hd.host} is not managed by that HMC. Actually managed "
            f"CPCs: {', '.join(cpc_names)}")

        cpc = client.cpcs.find(name=cpc_name)

        try:
            cpc.pull_full_properties()
        except zhmcclient.ConnectionError as exc:
            print(f"Cannot retrieve properties for CPC {cpc_name} "
                  f"(skipping it): {exc.__class__.__name__}: {exc}")
            continue

        cpc_props = dict(cpc.properties)
        for def_prop_name in def_cpc_props:

            if not is_cpc_property_hmc_inventory(def_prop_name):
                continue

            hmc_prop_name = def_prop_name.replace('_', '-')
            assert hmc_prop_name in cpc_props, (
                f"Property {def_prop_name!r} defined in inventory file for "
                f"CPC {cpc_name} for HMC {hd.nickname!r} does not actually "
                f"exist on that CPC (as {hmc_prop_name!r})")

            cpc_value = cpc_props[hmc_prop_name]
            def_cpc_value = def_cpc_props[def_prop_name]
            assert def_cpc_value == cpc_value, (
                f"Property {def_prop_name!r} defined in inventory file for "
                f"CPC {cpc_name} for HMC {hd.nickname!r} has an unexpected "
                f"value: inventory file: {def_cpc_value!r}, actual value: "
                f"{cpc_value!r}")


@pytest.mark.skip   # Disabled by default
def test_hmcdef_check_all_hmcs():
    """
    Check out the HMCs specified in the HMC inventory file.
    Skip HMCs that cannot be contacted.
    """
    hmcdefs = HMCDefinitions()
    hd_list = hmcdefs.list_all_hmcs()

    rt_config = zhmcclient.RetryTimeoutConfig(
        connect_timeout=10,
        connect_retries=1,
        read_timeout=30,
    )

    # Check real HMCs and their CPCs
    for hd in hd_list:

        if hd.mock_file:
            print(f"Skipping mocked HMC {hd.nickname!r}")
            continue

        for cpc_name in hd.cpcs:

            print(f"Checking CPC {cpc_name} defined for HMC {hd.nickname!r} at "
                  f"{hd.host}")

            session = zhmcclient.Session(
                hd.host, hd.userid, hd.password,
                verify_cert=hd.verify_cert,
                retry_timeout_config=rt_config)

            try:
                client = zhmcclient.Client(session)

                try:
                    session.logon()
                except zhmcclient.ConnectionError as exc:
                    print(f"Cannot logon to HMC at {hd.host} (skipping it): "
                          f"{exc.__class__.__name__}: {exc}")
                    continue

                cpcs = client.cpcs.list()
                cpc_names = [cpc.name for cpc in cpcs]

                assert cpc_name in cpc_names, (
                    f"CPC {cpc_name} defined in inventory file for HMC "
                    f"{hd.nickname!r} at {hd.host} is not managed by that "
                    f"HMC. Actually managed CPCs: {', '.join(cpc_names)}")

            finally:
                session.logoff()



# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for auto-updated resources (on CPCs in DPM mode).

These tests use partitions to test the auto-updating of resources. They do not
change any existing partitions, but create, modify and delete test partitions.
"""


import uuid
from time import sleep
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import
from zhmcclient_mock import FakedSession

from .utils import TEST_PREFIX, standard_partition_props, skip_warn

urllib3.disable_warnings()


def test_autoupdate_prop(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test auto-updated partitions when updating a property.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        if isinstance(cpc.manager.client.session, FakedSession):
            pytest.skip("Auto-update test requires notifications which are "
                        "not supported by zhmcclient_mock")

        print(f"Testing on CPC {cpc.name}")

        part_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"

        # Create the partition
        part_input_props = standard_partition_props(cpc, part_name)
        part = cpc.partitions.create(part_input_props)

        try:

            # Get a second zhmcclient object for the same partition, which will
            # be auto-updated
            part_auto = cpc.partitions.find(name=part_name)

            # Enable auto-update for the second partition
            part_auto.enable_auto_update()

            # Save some properties
            org_desc = part_auto.get_property('description')
            org_name = part_auto.name
            org_uri = part_auto.uri

            # Change the 'description' property through the first partition
            # object
            new_desc = org_desc + ' new'
            part.update_properties({'description': new_desc})

            # Test that the property change auto-updates the second partition
            # object.
            # We allow for some delay here, but in actual tests, the JMS message
            # about the change arrived faster than the operation response:
            #   07:20:59,817 Request Update Partition Properties
            #   07:21:00,529 JMS message for property change notification
            #   07:21:00,530 Response Update Partition Properties
            attempts = 20
            delay = 0.1  # seconds
            for _ in range(attempts):
                desc_auto = part_auto.properties['description']
                if desc_auto == new_desc:
                    break
                sleep(delay)
            assert desc_auto == new_desc, (
                f"Property did not auto-update after {attempts * delay} "
                "seconds")

            # Delete the partition through the first partition object
            part.delete()

            # Test that the second partition object is in ceased-existence
            # state.
            # We allow for some delay here, but in actual tests, the JMS message
            # about the change arrived faster than the operation response:
            #   07:21:00,532 Request Delete Partition
            #   07:21:02,735 JMS message for inventory change notification
            #   07:21:02,809 Response Delete Partition
            attempts = 20
            delay = 0.1  # seconds
            for _ in range(attempts):
                ceased = part_auto.ceased_existence
                if ceased:
                    break
                sleep(delay)
            assert ceased, (
                "Ceased-existence state did not auto-update after "
                f"{attempts * delay} seconds")

            # Test that accessing certain properties/methods on the
            # second (auto-updated) partition raises CeasedExistence

            with pytest.raises(zhmcclient.CeasedExistence) as exc_info:
                _ = part_auto.get_property('description')
            exc = exc_info.value
            assert exc.resource_uri == part.uri

            with pytest.raises(zhmcclient.CeasedExistence) as exc_info:
                _ = part_auto.prop('description')
            exc = exc_info.value
            assert exc.resource_uri == part.uri

            with pytest.raises(zhmcclient.CeasedExistence) as exc_info:
                part_auto.pull_full_properties()
            exc = exc_info.value
            assert exc.resource_uri == part.uri

            with pytest.raises(zhmcclient.CeasedExistence) as exc_info:
                _ = part_auto.dump()
            exc = exc_info.value
            assert exc.resource_uri == part.uri

            # Test that accessing certain properties/methods on the
            # second (auto-updated) partition does not raise CeasedExistence

            uri = part_auto.uri
            assert uri == org_uri

            name = part_auto.name
            assert name == org_name

            desc = part_auto.properties['description']
            assert desc == new_desc

            _ = part_auto.manager

            _ = part_auto.full_properties

            _ = part_auto.properties_timestamp

            ce = part_auto.ceased_existence
            assert ce is True

            _ = str(part_auto)

            _ = repr(part_auto)

            # Test that accessing properties of the first partition also
            # raises CeasedExistence (the ceased-existence state is set by
            # the delete() method).

            ce = part.ceased_existence
            assert ce is True

            with pytest.raises(zhmcclient.CeasedExistence) as exc_info:
                _ = part.get_property('description')
            exc = exc_info.value
            assert exc.resource_uri == part.uri

            with pytest.raises(zhmcclient.CeasedExistence) as exc_info:
                _ = part.prop('description')
            exc = exc_info.value
            assert exc.resource_uri == part.uri

            with pytest.raises(zhmcclient.CeasedExistence) as exc_info:
                part.pull_full_properties()
            exc = exc_info.value
            assert exc.resource_uri == part.uri

            with pytest.raises(zhmcclient.CeasedExistence) as exc_info:
                _ = part.dump()
            exc = exc_info.value
            assert exc.resource_uri == part.uri

        finally:
            # We want to make sure the test partition gets cleaned up after
            # the test, e.g. if the test is interrupted with Ctrl-C.
            try:
                part.delete()
            except zhmcclient.HTTPError as exc:
                # Since it normally will have been deleted already, we need to
                # allow for "not found".
                if exc.http_status == 404 and exc.reason == 1:
                    pass
                else:
                    raise


def test_autoupdate_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list() with auto-updated Partition manager.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        if isinstance(cpc.manager.client.session, FakedSession):
            pytest.skip("Auto-update test requires notifications which are "
                        "not supported by zhmcclient_mock")

        session = cpc.manager.session
        hd = session.hmc_definition

        new_part_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"

        # Get the initial set of partitions, for later comparison
        initial_part_list = cpc.partitions.list()
        if not initial_part_list:
            skip_warn(
                f"No partitions on CPC {cpc.name} managed by HMC {hd.host}")
        initial_part_names = {p.name for p in initial_part_list}

        # Enable auto-updating on partition manager and check partition list
        cpc.partitions.enable_auto_update()
        part_list = cpc.partitions.list()
        part_names = {p.name for p in part_list}
        assert part_names == initial_part_names

        try:

            # Create a partition and check partition list
            new_part_input_props = standard_partition_props(cpc, new_part_name)
            new_part = cpc.partitions.create(new_part_input_props)
            part_list = cpc.partitions.list()
            part_names = {p.name for p in part_list}
            exp_part_names = initial_part_names | {new_part.name}
            assert part_names == exp_part_names

            # Delete the partition and check partition list
            new_part.delete()
            part_list = cpc.partitions.list()
            part_names = {p.name for p in part_list}
            assert part_names == initial_part_names

            # Disable auto-updating on partition manager and check part. list
            cpc.partitions.disable_auto_update()
            part_list = cpc.partitions.list()
            part_names = {p.name for p in part_list}
            assert part_names == initial_part_names

            # Create a partition and check partition list
            new_part_input_props = standard_partition_props(cpc, new_part_name)
            new_part = cpc.partitions.create(new_part_input_props)
            part_list = cpc.partitions.list()
            part_names = {p.name for p in part_list}
            exp_part_names = initial_part_names | {new_part.name}
            assert part_names == exp_part_names

            # Delete the partition and check partition list
            new_part.delete()
            part_list = cpc.partitions.list()
            part_names = {p.name for p in part_list}
            assert part_names == initial_part_names

        finally:
            # We want to make sure the test partition gets cleaned up after
            # the test, e.g. if the test is interrupted with Ctrl-C.
            try:
                new_part.delete()
            except zhmcclient.HTTPError as exc:
                # Since it normally will have been deleted already, we need to
                # allow for "not found".
                if exc.http_status == 404 and exc.reason == 1:
                    pass
                else:
                    raise
